#######################################################################
#  TARDIS - Transformer And Rapid Dimensionless Instance Segmentation #
#                                                                     #
#  New York Structural Biology Center                                 #
#  Simons Machine Learning Center                                     #
#                                                                     #
#  Robert Kiewisz, Tristan Bepler                                     #
#  MIT License 2021 - 2024                                            #
#######################################################################
import sys
import time
from datetime import datetime
from os import listdir, getcwd
from os.path import isdir, isfile, join
from typing import Optional, Union
import platform

import numpy as np
import pandas as pd
import tifffile.tifffile as tif
import torch

from tardis_em.dist_pytorch.utils.utils import VoxelDownSampling

from tardis_em.dist_pytorch.datasets.patches import PatchDataSet
from tardis_em.dist_pytorch.dist import build_dist_network
from tardis_em.dist_pytorch.utils.build_point_cloud import BuildPointCloud
from tardis_em.dist_pytorch.utils.segment_point_cloud import PropGreedyGraphCut
from tardis_em.cnn.data_processing.draw_mask import (
    draw_semantic_membrane,
    draw_instances,
)
from tardis_em.cnn.data_processing.stitch import StitchImages
from tardis_em.cnn.data_processing.trim import trim_with_stride
from tardis_em.cnn.datasets.dataloader import PredictionDataset
from tardis_em.cnn.cnn import build_cnn_network
from tardis_em.cnn.data_processing.scaling import scale_image
from tardis_em.utils.aws import get_weights_aws, get_all_version_aws
from tardis_em.utils.device import get_device
from tardis_em.utils.errors import TardisError
from tardis_em.utils.export_data import NumpyToAmira, to_am, to_mrc, to_stl
from tardis_em.utils.load_data import load_am, ImportDataFromAmira, load_image
from tardis_em.utils.logo import print_progress_bar, TardisLogo
from tardis_em.utils.normalization import (
    MeanStdNormalize,
    RescaleNormalize,
    adaptive_threshold,
)
from tardis_em.utils.setup_envir import build_temp_dir, clean_up
from tardis_em.analysis.spatial_graph_utils import (
    FilterSpatialGraph,
    FilterConnectedNearSegments,
    SpatialGraphCompare,
    ComputeConfidenceScore,
)
from tardis_em.analysis.filament_utils import sort_by_length, resample_filament
from tardis_em.analysis.geometry_metrics import length_list
from tardis_em._version import version

# try:
#     from tardis_em.utils.ota_update import ota_update
#
#     ota = ota_update(status=True)
# except ImportError:
#     ota = ""
ota = ""

# Pytorch CUDA optimization
if torch.cuda.is_available():
    torch.backends.cudnn.benchmark = True


class GeneralPredictor:
    """
    MAIN WRAPPER FOR PREDICTION MT/MEM WITH TARDIS-PYTORCH

    Args:
        predict (str): Dataset type name.
        dir_ (str, np.ndarray): Dataset directory.
        output_format (str): Two output format for semantic and instance prediction.
        patch_size (int): Image 3D crop size.
        cnn_threshold (str): Threshold for CNN model.
        dist_threshold (float): Threshold for DIST model.
        points_in_patch (int): Maximum number of points per patched point cloud.
        predict_with_rotation (bool): If True, CNN predict with 4 90* rotations.
        amira_prefix (str): Optional, Amira file prefix used for spatial graph comparison.
        filter_by_length (float): Optional, filter setting for filtering short splines.
        connect_splines (int): Optional, filter setting for connecting near splines.
        connect_cylinder (int): Optional, filter setting for connecting splines
            withing cylinder radius.
        amira_compare_distance (int): Optional, compare setting, max distance between two splines
        to consider them as the same.
        amira_inter_probability (float): Optional, compare setting, portability threshold
        to define comparison class.
        instances (bool): If True, run instance segmentation after semantic.
        device_ (str): Define a computation device.
        debug (bool): If True, run in debugging mode.
    """

    def __init__(
        self,
        predict: str,
        dir_: Union[str, tuple[np.ndarray], np.ndarray],
        binary_mask: bool,
        output_format: str,
        patch_size: int,
        convolution_nn: str,
        cnn_threshold: str,
        dist_threshold: float,
        points_in_patch: int,
        predict_with_rotation: bool,
        instances: bool,
        device_: str,
        debug: bool,
        checkpoint: Optional[list] = None,
        model_version: Optional[int] = None,
        correct_px: float = None,
        normalize_px: float = None,
        amira_prefix: str = None,
        filter_by_length: int = None,
        connect_splines: int = None,
        connect_cylinder: int = None,
        amira_compare_distance: int = None,
        amira_inter_probability: float = None,
        tardis_logo: bool = True,
    ):
        self.transformation, self.px, self.image = None, None, None
        self.tardis_logo = tardis_logo
        self.tardis_progress = None
        self.title = None

        # Directories and dataset info
        self.dir = dir_
        self.output_format = output_format
        self.predict = predict
        if self.predict in ["Membrane2D", "Microtubule_tirf"]:
            self.expect_2d = True
        else:
            self.expect_2d = False
        self.amira_prefix = amira_prefix
        self.checkpoint = checkpoint
        self.model_version = model_version
        self.correct_px = correct_px
        self.normalize_px = normalize_px

        # Pre-processing setting
        self.cnn, self.dist = None, None
        self.patch_size = patch_size
        self.points_in_patch = points_in_patch
        self.pc_hd, self.pc_ld = np.zeros((0, 3)), np.zeros((0, 3))
        self.coords_df = np.zeros((1, 3))
        self.transformation = [0, 0, 0]
        self.segments, self.segments_filter = None, None

        # Prediction setting
        self.convolution_nn = convolution_nn
        self.cnn_threshold = cnn_threshold
        self.dist_threshold = dist_threshold
        self.rotate = predict_with_rotation

        # Global flags
        self.binary_mask = binary_mask
        self.predict_instance = instances
        self.device = get_device(device_)
        self.debug = debug
        self.semantic_header, self.instance_header, self.log_prediction = [], [], []

        """Initial Setup"""
        if debug:
            self.str_debug = " <Debugging Mode>"
        else:
            self.str_debug = ""

        # Check for the spatial graph in the folder from amira/tardis_em comp.
        self.amira_check = False
        if isinstance(self.dir, str):
            if isdir(join(self.dir, "amira")):
                self.amira_check = True
                self.dir_amira = join(dir_, "amira")

        # Searching for available images for prediction
        self.available_format = (
            ".tif",
            ".tiff",
            ".mrc",
            ".rec",
            ".am",
            ".map",
            ".npy",
        )
        self.omit_format = (
            "mask.tif",
            "mask.tiff",
            "mask.mrc",
            "mask.rec",
            "Correlation_Lines.am",
            "mask.am",
            "mask.map",
        )

        """Build handler's"""
        # Build handler's for reading data to correct format
        self.normalize = RescaleNormalize(clip_range=(1, 99))  # Normalize histogram
        self.mean_std = MeanStdNormalize()  # Standardize with mean and std

        # Sigmoid whole predicted image
        self.sigmoid = torch.nn.Sigmoid()

        # Build handler's for transforming data
        self.image_stitcher = StitchImages()
        self.post_processes = BuildPointCloud()

        # Build handler for DIST input and output
        if self.predict_instance:
            self.patch_pc = PatchDataSet(
                max_number_of_points=points_in_patch, graph=False
            )

            if predict in [
                "Actin",
                "Microtubule",
                "Membrane2D",
                "General_filament",
                "Microtubule_tirf",
            ]:
                self.GraphToSegment = PropGreedyGraphCut(
                    threshold=dist_threshold, connection=2, smooth=True
                )

                if predict == "Membrane2D":
                    self.filter_splines = FilterConnectedNearSegments(
                        distance_th=connect_splines,
                        cylinder_radius=connect_cylinder,
                    )
                else:
                    self.filter_splines = FilterSpatialGraph(
                        connect_seg_if_closer_then=connect_splines,
                        cylinder_radius=connect_cylinder,
                        filter_short_segments=(
                            filter_by_length if filter_by_length is not None else 0
                        ),
                    )
                self.compare_spline = SpatialGraphCompare(
                    distance_threshold=amira_compare_distance,
                    interaction_threshold=amira_inter_probability,
                )
                self.score_splines = ComputeConfidenceScore()
            elif predict in ["Membrane", "General_object"]:
                self.GraphToSegment = PropGreedyGraphCut(
                    threshold=dist_threshold, connection=8
                )

        # Build handler to output amira file
        self.create_headers()
        self.amira_file = NumpyToAmira(self.instance_header)

        # Sanity error checks
        self.init_check()

        # Build NN from checkpoints
        self.build_NN(NN=self.predict)

    def create_headers(self):
        dir_ = self.dir if isinstance(self.dir, str) else "np.ndarray"
        if dir_ == ".":
            dir_ = getcwd()

        main_ = [
            "# ASCII Spatial Graph",
            "# TARDIS - Transformer And Rapid Dimensionless Instance Segmentation (R)",
            f"# tardis_em-pytorch v{version} \r",
            f"# MIT License * 2021-{datetime.now().year} * "
            "Robert Kiewisz & Tristan Bepler",
            "Robert Kiewisz & Tristan Bepler",
            "",
            f"Directory: {dir_}",
            f"Output Format: {self.output_format}",
            f"Predict: {self.predict}",
            f"Device: {self.device}",
            "",
        ]

        if self.model_version is None:
            model = self.predict

            if model == "Actin":
                model = "Microtubule"
            elif model == "Microtubule":
                model = "microtubules_3d"
            elif model == "Microtubule_tirf":
                model = "microtubules_tirf"
            elif model == "Membrane":
                model = "membrane_3d"
            elif model == "Membrane2D":
                model = "membrane_2d"
            else:
                model = "None"

            if model != "None":
                if self.model_version is None:
                    model_version = get_all_version_aws(
                        self.convolution_nn, "32", model
                    )
                    model_version = model_version[-1]
                else:
                    model_version = self.model_version
            else:
                model_version = "None"

        self.semantic_header = [
            (
                f"CNN type: {self.convolution_nn} V_{model_version} "
                if self.checkpoint[0] is None
                else f"CNN model loaded from checkpoint: {self.checkpoint[0]}"
                f"Image patch size used for CNN: {self.patch_size}"
            ),
            f"CNN threshold: {self.cnn_threshold}",
            f"CNN predicted with 4x 90 degrees rotations: {self.rotate}",
        ]

        self.instance_header = [
            (
                "DIST 2D model"
                if self.predict
                in ["Actin", "Microtubule", "Membrane2D", "Microtubule_tirf"]
                else "DIST 3D model"
            ),
            (
                ""
                if self.checkpoint[1] is None
                else f"DIST model loaded from checkpoint: {self.checkpoint[1]}"
                f"DIST threshold: {self.dist_threshold}"
            ),
            f"Maximum number of points used in single DIST run: {self.points_in_patch}",
        ]

        self.log_prediction = (
            main_
            + ["----Semantic Segmentation----"]
            + self.semantic_header
            + ["", "----Instance Segmentation----"]
            + self.instance_header
            + [""]
        )

    def init_check(self):
        """
        All sanities check before TARDIS initialize prediction
        """
        msg = f"TARDIS v.{version} supports only MT and Mem segmentation!"
        assert_ = self.predict in [
            "Actin",
            "Membrane2D",
            "Membrane",
            "Microtubule",
            "Microtubule_tirf",
            "General_filament",
            "General_object",
        ]
        if self.tardis_logo:
            # Check if users ask to predict the correct structure
            if not assert_:
                TardisError(
                    id_="01",
                    py="tardis_em/utils/predictor.py",
                    desc=msg,
                )
                sys.exit()
        else:
            assert assert_, msg

        # Initiate log output
        if self.tardis_logo:
            self.tardis_progress = TardisLogo()

            if isdir(join(self.dir, "amira")):
                self.title = (
                    f"Fully-automatic Instance {self.predict} segmentation module "
                    f"with Amira comparison {self.str_debug}"
                )
            elif self.predict_instance:
                if self.output_format.startswith("None"):
                    self.title = (
                        f"Fully-automatic Instance {self.predict} segmentation module | {ota} "
                        f"{self.str_debug}"
                    )
                else:
                    self.title = (
                        f"Fully-automatic Semantic-Instance {self.predict} segmentation module | {ota} "
                        f"{self.str_debug}"
                    )
            else:
                self.title = (
                    f"Fully-automatic Semantic {self.predict} segmentation module | {ota} "
                    f"{self.str_debug}"
                )

            self.tardis_progress(title=self.title, text_2=f"Device: {self.device}")

        # Check for any other errors
        # Early stop if not semantic of instance was specified
        msg = f"Require that at lest one output format is not None but {self.output_format} was given!"
        assert_ = self.output_format == "None_None"
        if assert_:
            if self.tardis_logo:
                TardisError(
                    id_="151",
                    py="tardis_em/utils/predictor.py",
                    desc=msg,
                )
                sys.exit()
            else:
                assert assert_, msg

        # Check for know error if using ARM64 machine
        msg = f"STL output is not allowed on {platform.machine()} machine type!"
        assert_ = platform.machine() == "aarch64"
        if self.output_format.endswith("stl"):
            if self.tardis_logo:
                TardisError(
                    id_="151",
                    py="tardis_em/utils/predictor.py",
                    desc=msg,
                )
                sys.exit()
            else:
                assert not assert_, msg

    def build_NN(self, NN: str):
        if NN in ["Actin", "Microtubule", "Microtubule_tirf"]:
            self.normalize_px = 25 if self.normalize_px is None else self.normalize_px

            # Build CNN network with loaded pre-trained weights
            if not self.binary_mask:
                if NN in ["Actin", "Microtubule"]:
                    self.cnn = Predictor(
                        checkpoint=self.checkpoint[0],
                        network=self.convolution_nn,
                        subtype="32",
                        model_type=(
                            "microtubules_3d" if NN == "Microtubule" else "actin_3d"
                        ),
                        model_version=self.model_version,
                        img_size=self.patch_size,
                        sigmoid=False,
                        device=self.device,
                    )
                else:
                    self.cnn = Predictor(
                        checkpoint=self.checkpoint[0],
                        network=self.convolution_nn,
                        subtype="32",
                        model_type="microtubules_tirf",
                        model_version=self.model_version,
                        img_size=self.patch_size,
                        sigmoid=False,
                        _2d=True,
                        device=self.device,
                    )
            # Build DIST network with loaded pre-trained weights
            if not self.output_format.endswith("None"):
                self.dist = Predictor(
                    checkpoint=self.checkpoint[1],
                    network="dist",
                    subtype="triang",
                    model_type="2d",
                    model_version=self.model_version,
                    device=self.device,
                )
        elif NN in ["Membrane2D", "Membrane"]:
            self.normalize_px = 15 if self.normalize_px is None else self.normalize_px

            # Build CNN network with loaded pre-trained weights
            if NN == "Membrane2D":
                if not self.binary_mask:
                    self.cnn = Predictor(
                        checkpoint=self.checkpoint[0],
                        network=self.convolution_nn,
                        subtype="32",
                        model_type="membrane_2d",
                        model_version=self.model_version,
                        img_size=self.patch_size,
                        sigmoid=False,
                        device=self.device,
                        _2d=True,
                    )

                # Build DIST network with loaded pre-trained weights
                if not self.output_format.endswith("None"):
                    self.dist = Predictor(
                        checkpoint=self.checkpoint[1],
                        network="dist",
                        subtype="triang",
                        model_type="2d",
                        model_version=self.model_version,
                        device=self.device,
                    )
            else:
                if not self.binary_mask:
                    self.cnn = Predictor(
                        checkpoint=self.checkpoint[0],
                        network=self.convolution_nn,
                        subtype="32",
                        model_type="membrane_3d",
                        model_version=self.model_version,
                        img_size=self.patch_size,
                        sigmoid=False,
                        device=self.device,
                    )

                # Build DIST network with loaded pre-trained weights
                if not self.output_format.endswith("None"):
                    self.dist = Predictor(
                        checkpoint=self.checkpoint[1],
                        network="dist",
                        subtype="triang",
                        model_type="3d",
                        model_version=self.model_version,
                        device=self.device,
                    )
        elif NN.startswith("General"):
            self.cnn = Predictor(
                checkpoint=self.checkpoint[0],
                model_version=self.model_version,
                network=self.convolution_nn,
                img_size=self.patch_size,
                sigmoid=False,
                device=self.device,
                _2d=self.expect_2d,
            )
            if not self.output_format.endswith("None"):
                if NN.endswith("filament"):
                    self.dist = Predictor(
                        checkpoint=self.checkpoint[1],
                        network="dist",
                        subtype="triang",
                        model_type="2d",
                        model_version=self.model_version,
                        device=self.device,
                    )
                else:
                    self.dist = Predictor(
                        checkpoint=self.checkpoint[1],
                        network="dist",
                        subtype="triang",
                        model_type="3d",
                        model_version=self.model_version,
                        device=self.device,
                    )

    def load_data(self, id_name: Union[str, np.ndarray]):
        # Build temp dir
        build_temp_dir(dir_=self.dir)

        if isinstance(id_name, str):
            # Load image file
            if id_name.endswith(".am"):
                am = open(join(self.dir, id_name), "r", encoding="iso-8859-1").read(500)

                if "AmiraMesh 3D ASCII" in am:
                    self.amira_image = False
                    self.pc_hd = ImportDataFromAmira(
                        join(self.dir, id_name)
                    ).get_segmented_points()

                    self.image = None
                    self.px = self.correct_px
                    self.transformation = [0, 0, 0]

                    assert_ = self.pc_hd.shape[1] == 4
                    msg = f"Amira Spatial Graph has wrong dimension. Given {self.pc_hd.shape[1]}, but expected 4."
                    if self.tardis_logo:
                        if not assert_:
                            TardisError(
                                id_="11",
                                py="tardis_em/utils/predictor.py",
                                desc=msg,
                            )
                            sys.exit()
                    else:
                        assert assert_, msg
                else:
                    self.amira_image = True
                    self.image, self.px, _, self.transformation = load_am(
                        am_file=join(self.dir, id_name)
                    )
            else:
                self.amira_image = True
                self.image, self.px = load_image(join(self.dir, id_name))
                self.transformation = [0, 0, 0]
        else:
            self.image = id_name
            self.px = self.correct_px
            self.transformation = [0, 0, 0]

        # Normalize image histogram
        msg = f"Error while loading image {id_name}. Image loaded correctly, but output format "
        if self.amira_image and not self.binary_mask:
            self.image = self.normalize(self.mean_std(self.image)).astype(np.float32)

            # Sanity check image histogram
            if (
                not self.image.min() >= -1 or not self.image.max() <= 1
            ):  # Image not between in -1 and 1
                if self.image.min() >= 0 and self.image.max() <= 1:
                    self.image = (self.image - 0.5) * 2  # shift to -1 - 1
                elif self.image.min() >= 0 and self.image.max() <= 255:
                    self.image = self.image / 255  # move to 0 - 1
                    self.image = (self.image - 0.5) * 2  # shift to -1 - 1

            assert_ = self.image.dtype == np.float32
            if self.tardis_logo:
                if not assert_:
                    TardisError(
                        id_="11",
                        py="tardis_em/utils/predictor.py",
                        desc=msg + f" {self.image.dtype} is not float32!",
                    )
                    sys.exit()
            else:
                assert assert_, msg

            # Calculate parameters for image pixel size with optional scaling
            if self.correct_px is not None:
                self.px = self.correct_px

            if self.normalize_px is not None:
                self.scale_factor = self.px / self.normalize_px
            else:
                self.scale_factor = 1.0

            self.org_shape = self.image.shape
            self.scale_shape = np.multiply(self.org_shape, self.scale_factor).astype(
                np.int16
            )
            self.scale_shape = [int(i) for i in self.scale_shape]
        elif self.amira_image and self.binary_mask:
            # Check image structure
            self.image = np.where(self.image > 0, 1, 0).astype(np.int8)

            assert_ = self.image.dtype == np.int8 or self.image.dtype == np.uint8
            if self.tardis_logo:
                if not assert_:
                    TardisError(
                        id_="11",
                        py="tardis_em/utils/predictor.py",
                        desc=msg + f" {self.image.dtype} is not int8!",
                    )
                    sys.exit()
            else:
                assert assert_, msg

        if self.image.ndim == 2 or self.predict == "Microtubule_tirf":
            self.expect_2d = True
        else:
            self.expect_2d = False

    def predict_cnn(self, id_: int, id_name: str, dataloader):
        iter_time = 1
        if self.rotate:
            pred_title = f"CNN prediction with four 90 degree rotations with {self.convolution_nn}"
        else:
            pred_title = f"CNN prediction with {self.convolution_nn}"

        for j in range(len(dataloader)):
            if j % iter_time == 0 and self.tardis_logo:
                # Tardis progress bar update
                self.tardis_progress(
                    title=self.title,
                    text_1=f"Found {len(self.predict_list)} images to predict!",
                    text_2=f"Device: {self.device}",
                    text_3=f"Image {id_ + 1}/{len(self.predict_list)}: {id_name}",
                    text_4=f"Org. Pixel size: {self.px} A | Norm. Pixel size: {self.normalize_px}",
                    text_5=pred_title,
                    text_7="Current Task: CNN prediction...",
                    text_8=print_progress_bar(j, len(dataloader)),
                )

            # Pick image['s]
            input_, name = dataloader.__getitem__(j)

            if j == 0:
                start = time.time()

                # Predict
                input_ = self.cnn.predict(input_[None, :], rotate=self.rotate)

                # Scale progress bar refresh to 10s
                end = time.time()
                iter_time = 10 // (end - start)
                if iter_time <= 1:
                    iter_time = 1
            else:
                # Predict
                input_ = self.cnn.predict(input_[None, :], rotate=self.rotate)

            tif.imwrite(join(self.output, f"{name}.tif"), input_)

    def predict_cnn_napari(self, input_: torch.Tensor(), name: str):
        input_ = self.cnn.predict(input_[None, :], rotate=self.rotate)
        tif.imwrite(join(self.output, f"{name}.tif"), input_)

        return input_

    def postprocess_CNN(self, id_name: str):
        # Stitch predicted image patches
        if self.expect_2d:
            self.image = self.image_stitcher(
                image_dir=self.output, mask=False, dtype=np.float32
            )
            if self.image.ndim == 3:
                self.image = self.image[:, : self.scale_shape[0], : self.scale_shape[1]]
            else:
                self.image = self.image[: self.scale_shape[0], : self.scale_shape[1]]
        else:
            self.image = self.image_stitcher(
                image_dir=self.output, mask=False, dtype=np.float32
            )[: self.scale_shape[0], : self.scale_shape[1], : self.scale_shape[2]]

        # Restored original image pixel size
        self.image, _ = scale_image(image=self.image, scale=self.org_shape)
        self.image = torch.sigmoid(torch.from_numpy(self.image)).cpu().detach().numpy()

        # Threshold CNN prediction
        if self.cnn_threshold == "auto":
            self.image = adaptive_threshold(self.image)
        else:
            self.cnn_threshold = float(self.cnn_threshold)

            if self.cnn_threshold != 0:
                self.image = np.where(self.image >= self.cnn_threshold, 1, 0).astype(
                    np.uint8
                )
            else:
                if not self.output_format.startswith("return"):
                    self.log_prediction.append(
                        f"Semantic Prediction: {id_name[:-self.in_format]}"
                        f" | Number of pixels: {np.sum(self.image)}"
                    )
                    with open(join(self.am_output, "prediction_log.txt"), "w") as f:
                        f.write(" \n".join(self.log_prediction))

                    tif.imwrite(
                        join(self.am_output, f"{id_name[:-self.in_format]}_CNN.tif"),
                        self.image,
                    )
                    self.image = None

        """Clean-up temp dir"""
        clean_up(dir_=self.dir)

    def preprocess_DIST(self, id_name: str):
        if self.amira_image:
            # Post-process predicted image patches
            if self.predict in ["Actin", "Microtubule", "General_filament"]:
                self.pc_hd, self.pc_ld = self.post_processes.build_point_cloud(
                    image=self.image, down_sampling=5
                )
            elif self.predict in ["Membrane2D", "Microtubule_tirf"]:
                self.pc_hd, self.pc_ld = self.post_processes.build_point_cloud(
                    image=self.image, down_sampling=5, as_2d=True
                )
            else:
                self.pc_hd, self.pc_ld = self.post_processes.build_point_cloud(
                    image=self.image, down_sampling=5, as_2d=True
                )

            self.image = None
            self._debug(id_name=id_name, debug_id="pc")
        else:
            self.pc_hd = resample_filament(self.pc_hd, self.px)
            down_sample = VoxelDownSampling(voxel=5, labels=False, KNN=True)
            self.pc_ld = down_sample(coord=self.pc_hd[:, 1:])

    def predict_DIST(self, id_: int, id_name: str):
        iter_time = int(round(len(self.coords_df) / 10))

        if iter_time == 0:
            iter_time = 1
        if iter_time >= len(self.coords_df):
            iter_time = 10

        pc = self.pc_ld.shape
        graphs = []
        for id_dist, coord in enumerate(self.coords_df):
            if id_dist % iter_time == 0 and self.tardis_logo:
                self.tardis_progress(
                    title=self.title,
                    text_1=f"Found {len(self.predict_list)} images to predict!",
                    text_2=f"Device: {self.device}",
                    text_3=f"Image {id_ + 1}/{len(self.predict_list)}: {id_name}",
                    text_4=f"Org. Pixel size: {self.px} A | Norm. Pixel size: {self.normalize_px}",
                    text_5=f"Point Cloud: {pc[0]} Nodes; NaN Segments",
                    text_7="Current Task: DIST prediction...",
                    text_8=print_progress_bar(id_dist, len(self.coords_df)),
                )

            graph = self.dist.predict(x=coord[None, :])
            graphs.append(graph)

        # Tardis progress bar update
        if self.tardis_logo:
            self.tardis_progress(
                title=self.title,
                text_1=f"Found {len(self.predict_list)} images to predict!",
                text_2=f"Device: {self.device}",
                text_3=f"Image {id_ + 1}/{len(self.predict_list)}: {id_name}",
                text_4=f"Org. Pixel size: {self.px} A | Norm. Pixel size: {self.normalize_px}",
                text_5=f"Point Cloud: {self.pc_ld.shape[0]}; NaN Segments",
                text_7=f"Current Task: {self.predict} segmentation...",
            )

        return graphs

    def postprocess_DIST(self, id_, i):
        self.pc_ld = (
            self.pc_ld * self.px
            if self.correct_px is None
            else self.pc_ld * self.correct_px
        )
        self.pc_ld = self.pc_ld + self.transformation

        if self.predict in [
            "Actin",
            "Microtubule",
            "General_filament",
            "Microtubule_tirf",
        ]:
            self.log_tardis(id_, i, log_id=6.1)
        else:
            self.log_tardis(id_, i, log_id=6.2)

        if self.predict in [
            "Actin",
            "Microtubule",
            "Membrane2D",
            "General_filament",
            "Microtubule_tirf",
        ]:
            sort = True
            prune = 5
        else:
            sort = False
            prune = 15

        try:
            self.segments = self.GraphToSegment.patch_to_segment(
                graph=self.graphs,
                coord=self.pc_ld,
                idx=self.output_idx,
                sort=sort,
                prune=prune,
            )
            self.segments = sort_by_length(self.segments)
        except:
            self.segments = None

        # Tardis progress bar update
        if self.tardis_logo:
            no_segments = ""
            if self.segments is None:
                no_segments = f"Point Cloud: {self.pc_ld.shape[0]}; None Segments"
            else:
                no_ = np.max(np.unique(self.segments[:, 0])).item(0)
                no_segments = f"Point Cloud: {self.pc_ld.shape[0]}; {no_} Segments"

            self.tardis_progress(
                title=self.title,
                text_1=f"Found {len(self.predict_list)} images to predict!",
                text_2=f"Device: {self.device}",
                text_3=f"Image {id_ + 1}/{len(self.predict_list)}: {i}",
                text_4=f"Org. Pixel size: {self.px} A | Norm. Pixel size: {self.normalize_px}",
                text_5=no_segments,
                text_7=f"Current Task: {self.predict} segmented...",
            )

    def get_file_list(self):
        # Pickup files for the prediction
        if not isinstance(self.dir, str):
            if isinstance(self.dir, tuple) or isinstance(self.dir, list):
                self.predict_list = self.dir
            else:
                self.predict_list = [self.dir]
        else:
            if isdir(self.dir):
                self.predict_list = [
                    f
                    for f in listdir(self.dir)
                    if f.endswith(self.available_format)
                    and not f.endswith(self.omit_format)
                ]
            else:
                self.predict_list = [
                    f
                    for f in [self.dir]
                    if f.endswith(self.available_format)
                    and not f.endswith(self.omit_format)
                ]
        out_ = [
            i
            for i in self.dir.split("/")
            if not i.endswith((".mrc", ".rec", ".map", ".tif", ".tiff", ".am"))
        ]
        self.dir = join("/".join(out_))

        # Update Dir paths
        self.output = join(self.dir, "temp", "Predictions")
        self.am_output = join(self.dir, "Predictions")

        # Check if there is anything to predict in the user-indicated folder
        msg = f"Given {self.dir} does not contain any recognizable file!"
        assert_ = len(self.predict_list) == 0
        if self.tardis_logo:
            if assert_:
                TardisError(
                    id_="12",
                    py="tardis_em/utils/predictor.py",
                    desc=msg,
                )
                sys.exit()
            else:
                self.tardis_progress(
                    title=self.title,
                    text_1=f"Found {len(self.predict_list)} images to predict!",
                    text_2=f"Device: {self.device}",
                    text_7="Current Task: Setting-up environment...",
                )
        else:
            assert not assert_, msg

    def log_tardis(self, id_: int, i: Union[str, np.ndarray], log_id: float):
        if isinstance(i, np.ndarray):
            i = "Numpy array"

        if not self.tardis_logo:
            return

        # Common text for all configurations
        common_text = {
            "text_1": f"Found {len(self.predict_list)} images to predict!",
            "text_2": f"Device: {self.device}",
            "text_3": f"Image {id_ + 1}/{len(self.predict_list)}: {i}",
        }

        if log_id == 7:
            no_segments = np.max(self.segments[:, 0]) if len(self.segments) > 0 else 0
        else:
            no_segments = "None"

        # Define text configurations for each log_id
        pc_ld_shape = self.pc_ld.shape[0]

        text_configurations = {
            0: {
                **common_text,
                "text_4": "Org. Pixel size: Nan A",
                "text_7": "Current Task: Preprocessing for CNN...",
            },
            1: {
                **common_text,
                "text_4": f"Org. Pixel size: {self.px} A | Norm. Pixel size: {self.normalize_px}",
                "text_7": f"Current Task: Sub-dividing images for {self.patch_size} size",
            },
            2: {
                **common_text,
                "text_4": f"Org. Pixel size: {self.px} A | Norm. Pixel size: {self.normalize_px}",
                "text_7": "Current Task: Stitching...",
            },
            3: {
                **common_text,
                "text_4": f"Org. Pixel size: {self.px} A | Norm. Pixel size: {self.normalize_px}",
                "text_5": "Point Cloud: In processing...",
                "text_7": "Current Task: Image Postprocessing...",
            },
            4: {
                **common_text,
                "text_4": f"Org. Pixel size: {self.px} A | Norm. Pixel size: {self.normalize_px}",
                "text_5": f"Point Cloud: {pc_ld_shape} Nodes; NaN Segments",
                "text_7": "Current Task: Preparing for instance segmentation...",
            },
            5: {
                **common_text,
                "text_4": f"Org. Pixel size: {self.px} A | Norm. Pixel size: {self.normalize_px}",
                "text_5": f"Point Cloud: {pc_ld_shape} Nodes; NaN Segments",
                "text_7": "Current Task: DIST prediction...",
                "text_8": (
                    print_progress_bar(0, len(self.coords_df))
                    if self.tardis_logo
                    else ""
                ),
            },
            6.1: {
                **common_text,
                "text_4": f"Org. Pixel size: {self.px} A | Norm. Pixel size: {self.normalize_px}",
                "text_5": f"Point Cloud: {pc_ld_shape} Nodes; NaN Segments",
                "text_7": "Current Task: Instance Segmentation...",
                "text_8": "Filament segmentation is fitted to:",
                "text_9": f"pixel size: {self.px}; transformation: {self.transformation}",
            },
            6.2: {
                **common_text,
                "text_4": f"Org. Pixel size: {self.px} A | Norm. Pixel size: {self.normalize_px}",
                "text_5": f"Point Cloud: {pc_ld_shape}; NaN Segments",
                "text_7": "Current Task: Instance segmentation...",
            },
            7: {
                **common_text,
                "text_4": f"Org. Pixel size: {self.px} A | Norm. Pixel size: {self.normalize_px}",
                "text_5": f"Point Cloud: {pc_ld_shape} Nodes; {no_segments} Segments",
                "text_7": "Current Task: Segmentation finished!",
            },
        }

        # Retrieve text configuration based on log_id
        config = text_configurations.get(log_id)

        # Tardis progress bar update
        self.tardis_progress(title=self.title, **config)

    def save_semantic_mask(self, i):
        self.log_prediction.append(
            f"Semantic Prediction: {i[:-self.in_format]}"
            f" | Number of pixels: {np.sum(self.image)}"
        )

        with open(join(self.am_output, "prediction_log.txt"), "w") as f:
            f.write(" \n".join(self.log_prediction))

        if self.output_format.startswith("mrc"):
            to_mrc(
                data=self.image,
                file_dir=join(self.am_output, f"{i[:-self.in_format]}_semantic.mrc"),
                pixel_size=(self.px if self.correct_px is None else self.correct_px),
                label=self.semantic_header,
            )
        elif self.output_format.startswith("tif"):
            tif.imwrite(
                join(self.am_output, f"{i[:-self.in_format]}_semantic.tif"),
                (
                    np.flip(self.image, 1)
                    if i.endswith((".mrc", ".rec"))
                    else self.image
                ),
            )
        elif self.output_format.startswith("am"):
            to_am(
                data=self.image,
                file_dir=join(self.am_output, f"{i[:-self.in_format]}_semantic.am"),
                pixel_size=(self.px if self.correct_px is None else self.correct_px),
                header=self.semantic_header,
            )
        elif self.output_format.startswith("npy"):
            np.save(
                join(self.am_output, f"{i[:-self.in_format]}_semantic.npy"), self.image
            )

    def save_instance_PC(self, i):
        self.log_prediction.append(
            f"Instance Prediction: {i[:-self.in_format]}; Number of segments: {np.max(self.segments[:, 0])}"
        )

        if self.predict in [
            "Actin",
            "Microtubule",
            "General_filament",
            "Microtubule_tirf",
        ]:
            self.segments_filter = self.filter_splines(segments=self.segments)
            self.segments_filter = sort_by_length(self.segments_filter)

            self.log_prediction.append(
                f"Instance Prediction: {i[:-self.in_format]};"
                f" Number of segments: {np.max(self.segments_filter[:, 0])}"
            )

        with open(join(self.am_output, "prediction_log.txt"), "w") as f:
            f.write(" \n".join(self.log_prediction))

        if self.output_format.endswith("amSG") and self.predict in [
            "Actin",
            "Microtubule",
            "General_filament",
            "Microtubule_tirf",
        ]:
            self.amira_file.export_amira(
                coords=self.segments,
                file_dir=join(self.am_output, f"{i[:-self.in_format]}_SpatialGraph.am"),
                labels=["TardisPrediction"],
                scores=[
                    ["EdgeLength", "EdgeConfidenceScore"],
                    [
                        length_list(self.segments),
                        self.score_splines(self.segments),
                    ],
                ],
            )

            self.amira_file.export_amira(
                coords=self.segments_filter,
                file_dir=join(
                    self.am_output,
                    f"{i[:-self.in_format]}_SpatialGraph_filter.am",
                ),
                labels=["TardisPrediction"],
                scores=[
                    ["EdgeLength", "EdgeConfidenceScore"],
                    [
                        length_list(self.segments_filter),
                        self.score_splines(self.segments_filter),
                    ],
                ],
            )

            if self.amira_check and self.predict == "Microtubule":
                dir_amira_file = join(
                    self.dir_amira,
                    i[: -self.in_format] + self.amira_prefix + ".am",
                )

                if isfile(dir_amira_file):
                    amira_sg = ImportDataFromAmira(src_am=dir_amira_file)
                    amira_sg = amira_sg.get_segmented_points()

                    if amira_sg is not None:
                        compare_sg, label_sg = self.compare_spline(
                            amira_sg=amira_sg, tardis_sg=self.segments_filter
                        )

                        if self.output_format.endswith("amSG"):
                            self.amira_file.export_amira(
                                file_dir=join(
                                    self.am_output,
                                    f"{i[:-self.in_format]}_AmiraCompare.am",
                                ),
                                coords=compare_sg,
                                labels=label_sg,
                                scores=None,
                            )
        elif self.output_format.endswith("csv"):
            segments = pd.DataFrame(self.segments)
            segments.to_csv(
                join(self.am_output, f"{i[:-self.in_format]}_instances.csv"),
                header=["IDs", "X [A]", "Y [A]", "Z [A]"],
                index=False,
                sep=",",
            )

            if self.predict in [
                "Actin",
                "Microtubule",
                "Microtubule_tirf",
                "Membrane2D",
                "General_filament",
            ]:
                self.segments_filter = pd.DataFrame(self.segments_filter)
                self.segments_filter.to_csv(
                    join(
                        self.am_output,
                        f"{i[:-self.in_format]}_instances_filter.csv",
                    ),
                    header=["IDs", "X [A]", "Y [A]", "Z [A]"],
                    index=False,
                    sep=",",
                )
                self.segments_filter = self.segments_filter.to_numpy()
        elif self.output_format.endswith(("mrc", "tif", "am")):
            if self.predict in [
                "Membrane",
                "Membrane2D",
                "General_object",
                "Microtubule_tirf",
            ]:
                self.mask_semantic = draw_semantic_membrane(
                    mask_size=self.org_shape,
                    coordinate=self.segments,
                    pixel_size=(
                        self.px if self.correct_px is None else self.correct_px
                    ),
                    spline_size=(
                        60
                        if self.predict
                        not in [
                            "Microtubule_tirf",
                            "Microtubule",
                            "General_filament",
                            "Actin",
                        ]
                        else 25
                    ),
                )
            else:
                self.segments[:, 1:] = (
                    self.segments[:, 1:] / self.px
                    if self.correct_px is None
                    else self.correct_px
                )

                self.mask_semantic = draw_instances(
                    mask_size=self.org_shape,
                    coordinate=self.segments,
                    pixel_size=(
                        self.px if self.correct_px is None else self.correct_px
                    ),
                    circle_size=125,
                )
            self._debug(id_name=i, debug_id="instance_mask")

            if self.output_format.endswith("mrc"):
                to_mrc(
                    data=self.mask_semantic,
                    file_dir=join(
                        self.am_output, f"{i[:-self.in_format]}_instance.mrc"
                    ),
                    pixel_size=(
                        self.px if self.correct_px is None else self.correct_px
                    ),
                    label=self.instance_header,
                )
            elif self.output_format.endswith("tif"):
                tif.imwrite(
                    join(self.am_output, f"{i[:-self.in_format]}_instance.tif"),
                    self.mask_semantic,
                )
            elif self.output_format.endswith("am"):
                to_am(
                    data=self.mask_semantic,
                    file_dir=join(self.am_output, f"{i[:-self.in_format]}_instance.am"),
                    pixel_size=(
                        self.px if self.correct_px is None else self.correct_px
                    ),
                    header=self.instance_header,
                )
        elif self.output_format.endswith("stl"):
            if self.predict == "Membrane":
                to_stl(
                    data=self.segments,
                    file_dir=join(self.am_output, f"{i[:-self.in_format]}.stl"),
                )
        elif self.output_format.endswith("npy"):
            np.save(
                join(self.am_output, f"{i[:-self.in_format]}_instance.npy"),
                self.segments,
            )

    def _debug(self, id_name: str, debug_id: str):
        if self.debug:
            if debug_id == "cnn":
                tif.imwrite(
                    join(self.am_output, f"{id_name[:-self.in_format]}_CNN.tif"),
                    self.image,
                )
            elif debug_id == "pc":
                np.save(
                    join(self.am_output, f"{id_name[:-self.in_format]}_raw_pc_hd.npy"),
                    self.pc_hd,
                )
                np.save(
                    join(self.am_output, f"{id_name[:-self.in_format]}_raw_pc_ld.npy"),
                    self.pc_ld,
                )
            elif debug_id == "graph":
                try:
                    np.save(
                        join(
                            self.am_output,
                            f"{id_name[:-self.in_format]}_graph_voxel.npy",
                        ),
                        self.graphs,
                    )
                except ValueError:
                    pass
            elif debug_id == "segment":
                if self.device == "cpu":
                    np.save(
                        join(
                            self.am_output,
                            f"{id_name[:-self.in_format]}_coord_voxel.npy",
                        ),
                        self.coords_df,
                    )
                    np.save(
                        join(
                            self.am_output, f"{id_name[:-self.in_format]}_idx_voxel.npy"
                        ),
                        self.output_idx,
                    )
                else:
                    try:
                        np.save(
                            join(
                                self.am_output,
                                f"{id_name[:-self.in_format]}_coord_voxel.npy",
                            ),
                            self.coords_df.cpu().detach().numpy(),
                        )
                    except AttributeError:
                        pass
                    try:
                        np.save(
                            join(
                                self.am_output,
                                f"{id_name[:-self.in_format]}_idx_voxel.npy",
                            ),
                            self.output_idx.cpu().detach().numpy(),
                        )
                    except:
                        pass
                np.save(
                    join(self.am_output, f"{id_name[:-self.in_format]}_segments.npy"),
                    self.segments,
                )
            elif debug_id == "instance_mask":
                np.save(
                    join(
                        self.am_output, f"{id_name[:-self.in_format]}_instance_mask.npy"
                    ),
                    self.mask_semantic,
                )

    def __call__(self):
        """Process each image with CNN and DIST"""
        self.get_file_list()

        semantic_output, instance_output, instance_filter_output = [], [], []
        for id_, i in enumerate(self.predict_list):
            """CNN Pre-Processing"""
            if isinstance(i, str):
                # Find a file format
                self.in_format = 0
                if i.endswith((".tif", ".mrc", ".rec", ".map", ".npy")):
                    self.in_format = 4
                elif i.endswith(".tiff"):
                    self.in_format = 5
                elif i.endswith(".am"):
                    self.in_format = 3

            # Tardis progress bar update
            self.log_tardis(id_, i, log_id=0)

            # Load data
            self.load_data(id_name=i)

            msg = (
                f"Predicted file {id_} is numpy array without pixel size metadate {self.px}."
                "Please pass correct_px argument as a correct pixel size value."
            )
            assert_ = self.px is None and not isinstance(i, str)
            if assert_:
                if self.tardis_logo:
                    TardisError(id_="161", py="tardis_em.utils.predictor.py", desc=msg)
                    sys.exit()
                else:
                    assert not assert_, msg

            # Tardis progress bar update
            self.log_tardis(id_, i, log_id=1)

            """Semantic Segmentation"""
            if not self.binary_mask:
                # Cut image for fix patch size and normalizing image pixel size
                trim_with_stride(
                    image=self.image,
                    scale=self.scale_shape,
                    trim_size_xy=self.patch_size,
                    trim_size_z=self.patch_size,
                    output=join(self.dir, "temp", "Patches"),
                    image_counter=0,
                    clean_empty=False,
                    stride=round(self.patch_size * 0.125),
                )
                self.image = None

                """CNN prediction"""
                self.predict_cnn(
                    id_=id_,
                    id_name=i,
                    dataloader=PredictionDataset(
                        join(self.dir, "temp", "Patches", "imgs")
                    ),
                )

                """CNN Post-Processing"""
                # Tardis progress bar update
                self.log_tardis(id_, i, log_id=2)

                # Post-process
                self.postprocess_CNN(id_name=i)

                # Store optionally for return
                if self.output_format.startswith("return"):
                    semantic_output.append(self.image)

                # Check if the image is binary
                if self.image is not None and not self.output_format.startswith(
                    "return"
                ):
                    if len(pd.unique(self.image.flatten())) == 1:
                        self.image = None

                if self.image is None:
                    continue

                # Debug flag
                self._debug(id_name=i, debug_id="cnn")

                # Check if predicted image
                assert_ = self.image.shape == self.org_shape
                if not assert_:
                    msg = (
                        "Last Task: Stitching/Scaling/Make correction..."
                        f"Tardis Error: Error while converting to {self.px} A "
                        f"Org. shape {self.org_shape} is not the same as "
                        f"converted shape {self.image.shape}"
                    )
                    if self.tardis_logo:
                        TardisError(
                            id_="116",
                            py="tardis_em/utils/predictor.py",
                            desc=msg,
                        )
                    else:
                        assert assert_, msg
                    sys.exit()

                # Save predicted mask as file
                self.save_semantic_mask(i)

                # Sanity check for binary mask
                if not self.image.min() == 0 and not self.image.max() == 1:
                    continue

                # Close iter loop if not instance prediction
                if not self.predict_instance:
                    continue

            """Instance Segmentation"""
            # Tardis progress bar update
            self.log_tardis(id_, i, log_id=3)

            self.preprocess_DIST(id_name=i)
            self.segments = np.zeros((0, 4))
            self.segments_filter = None

            # Break iter loop for instances if no point cloud is found
            if len(self.pc_ld) == 0:
                if self.output_format.endswith("return"):
                    instance_output.append(np.zeros((0, 4)))
                    instance_filter_output.append(np.zeros((0, 4)))
                continue

            # Tardis progress bar update
            self.log_tardis(id_, i, log_id=4)

            # Build patches dataset
            (
                self.coords_df,
                _,
                self.output_idx,
                _,
            ) = self.patch_pc.patched_dataset(coord=self.pc_ld)

            # Predict point cloud
            self.log_tardis(id_, i, log_id=5)

            # DIST prediction
            self.graphs = self.predict_DIST(id_=id_, id_name=i)
            self._debug(id_name=i, debug_id="graph")
            # Save debugging check point
            self._debug(id_name=i, debug_id="segment")

            # DIST Instance graph-cut
            self.postprocess_DIST(id_, i)

            if self.segments is None:
                continue

            self.log_tardis(id_, i, log_id=7)

            """Save as .am"""
            self.save_instance_PC(i)

            if self.segments is None:
                if self.output_format.endswith("return"):
                    instance_output.append(np.zeros((0, 4)))
                    instance_filter_output.append(np.zeros((0, 4)))
            else:
                instance_output.append(self.segments)
                if self.predict in [
                    "Actin",
                    "Microtubule",
                    "General_filament",
                    "Microtubule_tirf",
                ]:
                    instance_filter_output.append(self.segments_filter)

            """Clean-up temp dir"""
            clean_up(dir_=self.dir)

        """Optional return"""
        if self.output_format.startswith("return"):
            if self.output_format.endswith("return"):
                return semantic_output, instance_output, instance_filter_output
            return semantic_output


class Predictor:
    """
    WRAPPER FOR PREDICTION

     Args:
         device (torch.device): Device on which to predict.
         checkpoint (str, Optional): Local weights files.
         network (str, Optional): Optional network type name.
         subtype (str, Optional): Optional model subtype name.
         model_type (str, Optional): Optional model type name.
         model_version (int, Optional): Optional model version.
         img_size (int, Optional): Optional image patch size.
         sigmoid (bool): Predict output with sigmoid.
    """

    def __init__(
        self,
        device: torch.device,
        network: Optional[str] = None,
        checkpoint: Optional[str] = None,
        subtype: Optional[str] = None,
        model_version: Optional[int] = None,
        img_size: Optional[int] = None,
        model_type: Optional[str] = None,
        sigma: Optional[float] = None,
        sigmoid=True,
        _2d=False,
        logo=True,
    ):
        self.logo = logo

        self.device = device
        self.img_size = img_size

        msg = "Missing network weights or network name!"
        if self.logo:
            if checkpoint is None and network is None:
                TardisError(
                    "139",
                    "tardis_em/utils/predictor.py",
                    msg,
                )
        else:
            assert checkpoint is not None and network is not None, msg

        if checkpoint is None:
            print(
                f"Searching for weight file for {network}_{subtype} for {model_type}..."
            )

            weights = torch.load(
                get_weights_aws(network, subtype, model_type, model_version),
                map_location="cpu",
                weights_only=False,
            )
        elif isinstance(checkpoint, dict):
            print("Loading weight dictionary...")
            weights = checkpoint
        else:
            print("Loading weight model...")
            weights = torch.load(checkpoint, map_location="cpu", weights_only=False)

        # Load model
        if isinstance(weights, dict):  # Deprecated support for dictionary
            # Allow overwriting sigma
            if sigma is not None:
                weights["model_struct_dict"]["coord_embed_sigma"] = sigma

            if "dist_type" in weights["model_struct_dict"]:
                from tardis_em.dist_pytorch.utils.utils import check_model_dict
            else:
                from tardis_em.cnn.utils.utils import check_model_dict

            model_structure = check_model_dict(weights["model_struct_dict"])

            if network is not None:
                self.network = network
            else:
                if "dist_type" in model_structure:
                    self.network = "dist"
                else:
                    self.network = "cnn"

            self._2d = _2d
            self.model = self._build_model_from_checkpoint(
                structure=model_structure, sigmoid=sigmoid
            )
            self.model.load_state_dict(weights["model_state_dict"])
        else:  # Load onnx or another model
            self.network = network
            self._2d = _2d
            self.model = weights

            if not network == "dist":
                self.model.update_patch_size(self.img_size, sigmoid)
        self.model.to(self.device)
        self.model.eval()

        del weights  # Cleanup weight file from memory

    def _build_model_from_checkpoint(self, structure: dict, sigmoid=True):
        """
        Use checkpoint metadata to build a compatible network

        Args:
            structure (dict): Metadata dictionary with network setting.
            sigmoid (bool): Predict output with sigmoid.

        Returns:
            pytorch model: NN pytorch model.
        """
        if "dist_type" in structure:
            model = build_dist_network(
                network_type=structure["dist_type"],
                structure=structure,
                prediction=sigmoid,
            )
        elif "cnn_type" in structure:
            model = build_cnn_network(
                network_type=structure["cnn_type"],
                structure=structure,
                img_size=self.img_size,
                prediction=sigmoid,
            )
        else:
            model = None

        return model

    def predict(
        self, x: torch.Tensor, y: Optional[torch.Tensor] = None, rotate=False
    ) -> np.ndarray:
        """
        General predictor.

        Args:
            x (torch.Tensor): Main feature used for prediction.
            y (torch.Tensor, None): Optional feature used for prediction.
            rotate (bool): Optional flag for CNN to output avg. From 4x 90* rotation

        Returns:
            np.ndarray: Predicted features.
        """
        if isinstance(x, np.ndarray):
            x = torch.Tensor(x)
        if y is not None and isinstance(y, np.ndarray):
            y = torch.Tensor(y)

        with torch.no_grad():
            dim_ = x.shape[-1]
            x = x.to(self.device)

            if self.network == "dist":
                if y is None:
                    out = self.model(coords=x, node_features=None)
                else:
                    out = self.model(coords=x, node_features=y.to(self.device))

                out = out.cpu().detach().numpy()[0, 0, :]
                g_len = out.shape[0]
                g_range = range(g_len)

                # Overwrite diagonal with 1
                out[g_range, g_range] = np.eye(g_len, g_len)[g_range, g_range]
                return out
            else:
                if rotate:
                    if self._2d:
                        out = np.zeros((dim_, dim_), dtype=np.float32)
                        for k in range(4):
                            x_ = torch.rot90(x, k=k, dims=(2, 3))
                            x_ = self.model(x_) / 4
                            x_ = x_.cpu().detach().numpy()[0, 0, :]

                            out += np.rot90(x_, k=-k, axes=(0, 1))
                    else:
                        out = np.zeros((dim_, dim_, dim_), dtype=np.float32)
                        for k in range(4):
                            x_ = torch.rot90(x, k=k, dims=(3, 4))
                            x_ = self.model(x_) / 4
                            x_ = x_.cpu().detach().numpy()[0, 0, :]

                            out += np.rot90(x_, k=-k, axes=(1, 2))
                else:
                    out = self.model(x).cpu().detach().numpy()[0, 0, :]

                return out



#######################################################################
#  TARDIS - Transformer And Rapid Dimensionless Instance Segmentation #
#                                                                     #
#  New York Structural Biology Center                                 #
#  Simons Machine Learning Center                                     #
#                                                                     #
#  Robert Kiewisz, Tristan Bepler                                     #
#  MIT License 2021 - 2024                                            #
#######################################################################
import torch
import torch.nn as nn


class SparsTriangularUpdate(nn.Module):
    """
    Module for updating a sparse tensor in a triangular fashion.

    This class applies a sequence of transformations, including normalization,
    sigmoid activation, linear transformations, and a triangular multiplication update.
    The transformations are designed to update the input tensor while preserving its sparsity.
    """

    def __init__(self, input_dim: int, channel_dim=128, axis=1, knn=8):
        """
        Initializes the SparsTriangularUpdate.

        Args:
            input_dim (int): The dimensionality of the input data.
            channel_dim (int): The number of channels in the output from some transformations.
            axis (int): The axis to be used in the triangular update rule.
        """
        super().__init__()
        self.input_dim = input_dim
        self.channel_dim = channel_dim
        self.axis = axis
        self.init_scaling = 1 / 1.4142135623730951
        self.knn = knn

        # Define the transformations to be applied
        self.norm_input = nn.LayerNorm(input_dim)

        self.linear_a = nn.Linear(input_dim, channel_dim)
        self.gate_a = nn.Linear(input_dim, channel_dim)

        self.linear_b = nn.Linear(input_dim, channel_dim)
        self.gate_b = nn.Linear(input_dim, channel_dim)

        self.norm_o = nn.LayerNorm(channel_dim)
        self.gate_o = nn.Linear(input_dim, input_dim)
        self.linear_o = nn.Linear(channel_dim, input_dim)

        self._reset_parameters()

    def _reset_parameters(self):
        """
        Initial parameter and bias scaling.
        """
        nn.init.xavier_uniform_(self.linear_a.weight, gain=self.init_scaling)
        nn.init.constant_(self.linear_a.bias, 0.0)

        nn.init.xavier_uniform_(self.linear_b.weight, gain=self.init_scaling)
        nn.init.constant_(self.linear_b.bias, 0.0)

        nn.init.constant_(self.linear_o.weight, 0.0)
        nn.init.constant_(self.linear_o.bias, 0.0)

    def forward(self, x: torch.tensor, indices: list) -> torch.tensor:
        """
        Forward pass for SparsTriangularUpdate.

        Args:
            x (torch.tensor): A sparse coordinate tensor containing the input data.
            indices (list): List of all indices

        Returns:
            torch.tensor, list: A sparse coordinate tensor representing the updated tensor.
        """
        x = self.norm_input(x)  # Batch x Length x Channels [N, Ch]

        # Compute intermediate transformations
        a = torch.sigmoid(self.gate_a(x)) * self.linear_a(x)
        b = torch.sigmoid(self.gate_b(x)) * self.linear_b(x)

        # Apply triangular multiplication update

        if self.axis == 1:  # Row-wise update
            k = torch.einsum("ik,ijk->ik", a, b[indices[0]])
        else:  # Col-wise update
            k = torch.einsum("ik,ijk->ik", a, b[indices[1]])

        return torch.sigmoid(self.gate_o(x)) * self.linear_o(self.norm_o(k))



#######################################################################
#  TARDIS - Transformer And Rapid Dimensionless Instance Segmentation #
#                                                                     #
#  New York Structural Biology Center                                 #
#  Simons Machine Learning Center                                     #
#                                                                     #
#  Robert Kiewisz, Tristan Bepler                                     #
#  MIT License 2021 - 2024                                            #
#######################################################################
from typing import Optional, Tuple, Union

import numpy as np
from scipy.interpolate import splev, splprep
from scipy.spatial.distance import cdist
from sklearn.neighbors import KDTree

from tardis_em.analysis.geometry_metrics import (
    angle_between_vectors,
    total_length,
    tortuosity,
)
from tardis_em.dist_pytorch.utils.utils import pc_median_dist


def resample_filament(points, spacing_size):
    """
    Resamples points for each filament so they have the given spacing size.

    Parameters:
    points (np.array): Array of shape [N, 4] where each column is [ID, X, Y, Z].
    spacing_size (float): Desired spacing size between points.

    Returns:
    np.array: Resampled array with the same structure.
    """
    unique_ids = np.unique(points[:, 0])
    resampled_points = []

    for filament_id in unique_ids:
        filament_points = points[points[:, 0] == filament_id][:, 1:]
        if filament_points.shape[0] < 2:
            resampled_points.append(points[points[:, 0] == filament_id])
            continue

        cumulative_distances = np.cumsum(
            np.sqrt(np.sum(np.diff(filament_points, axis=0) ** 2, axis=1))
        )
        cumulative_distances = np.insert(cumulative_distances, 0, 0)

        num_new_points = int(cumulative_distances[-1] // spacing_size)
        new_points = [filament_points[0]]

        for i in range(1, num_new_points + 1):
            target_distance = i * spacing_size
            idx = np.searchsorted(cumulative_distances, target_distance)
            if idx >= len(cumulative_distances):
                continue

            if cumulative_distances[idx] == target_distance:
                new_points.append(filament_points[idx])
            else:
                ratio = (target_distance - cumulative_distances[idx - 1]) / (
                    cumulative_distances[idx] - cumulative_distances[idx - 1]
                )
                new_point = filament_points[idx - 1] + ratio * (
                    filament_points[idx] - filament_points[idx - 1]
                )
                new_points.append(new_point)

        new_points = np.array(new_points)
        filament_ids = np.full((new_points.shape[0], 1), filament_id)
        resampled_points.append(np.hstack((filament_ids, new_points)))

    return np.vstack(resampled_points)


def sort_segments(coord: np.ndarray) -> np.ndarray:
    df = np.unique(coord[:, 0])

    new_coord = []
    for i in df:
        c = coord[np.isin(coord[:, 0], [i]), 1:]
        c = sort_segment(c)

        id_ = np.repeat(i, len(c))
        c = np.array((id_, c[:, 0], c[:, 1], c[:, 2])).T

        new_coord.append(c)

    return np.concatenate(new_coord)


def sort_segment(coord: np.ndarray) -> np.ndarray:
    """
    Sorting of the point cloud based on number of connections followed by
    searching of the closest point with cdist.

    Args:
        coord (np.ndarray): Coordinates for each unsorted point idx.

    Returns:
        np.ndarray: Array of point in line order.
    """
    if len(coord) < 2:
        return coord

    new_c = []
    for i in range(len(coord) - 1):
        if i == 0:
            id = np.where(
                [sum(i) for i in cdist(coord, coord)]
                == max([sum(i) for i in cdist(coord, coord)])
            )[0]

            new_c.append(coord[id[0]])
            coord = np.delete(coord, id[0], 0)

        kd = KDTree(coord)
        points = kd.query(np.expand_dims(new_c[len(new_c) - 1], 0))[1][0][0]

        new_c.append(coord[points])
        coord = np.delete(coord, points, 0)

    if len(new_c) > 0:
        return np.stack(new_c)
    else:
        return coord


def reorder_segments_id(
    coord: np.ndarray,
    order_range: Optional[list] = None,
    order_list: Optional[Union[list, np.ndarray]] = None,
) -> np.ndarray:
    """
    Reorder list of segments to remove missing IDs

    E.g. Change IDs from [1, 2, 3, 5, 6, 8] to [1, 2, 3, 4, 5, 6]

    Args:
        coord: Array of points in 3D or 3D with their segment ID
        order_range: Costume id range for reordering
        order_list: List of reorder IDs to match to coord.

    Returns:
        np.ndarray: Array of points with reordered IDs values
    """
    df = np.unique(coord[:, 0])

    if order_range is None:
        df_range = np.asarray(range(0, len(df)), dtype=df.dtype)
    else:
        df_range = np.asarray(range(order_range[0], order_range[1]), dtype=df.dtype)

    if order_list is None:
        for id, i in enumerate(coord[:, 0]):
            coord[id, 0] = df_range[np.where(df == i)[0][0]]
    else:
        ordered_coord = []
        for i, new_id in zip(df, order_list):
            line = coord[np.where(coord[:, 0] == i)[0], :]
            line[:, 0] = new_id
            ordered_coord.append(line)
        coord = np.concatenate(ordered_coord)
        coord = reorder_segments_id(coord)

    return coord


def smooth_spline(points: np.ndarray, s=0.5):
    """
    Spline smoothing given an 's' smoothness factor.

    Args:
        points (np.ndarray): Point array [(ID), X, Y, Z] with optional ID and Z
        dimension.
        s (float): Smoothness factor.

    Returns:
        Returns: Smooth spline
    """
    if points.shape[1] == 4:  # [ID, X, Y, Z]
        id_ = int(points[0, 0])
        points = points[:, 1:]
        norm_pc = pc_median_dist(points)
        points = points / norm_pc

        t_before = tortuosity(points)
        try:
            tck, u = splprep(points.T, s=s)
            spline = np.stack(splev(u, tck)).T
        except ValueError:
            spline = points

        spline = spline * norm_pc
        t_after = tortuosity(spline)
        ids = np.zeros((len(spline), 1))
        ids[:, 0] = id_

        # Sanity check if spline smoothing failed
        if t_after > t_before:
            return np.hstack((ids, points * norm_pc))
        return np.hstack((ids, spline))
    else:  # [X, Y, Z]
        tck, u = splprep(points.T)

        return np.stack(splev(u, tck)).T


def sort_by_length(coord):
    """
    Sort all splines by their length.

    Args:
        coord: Array of coordinates.

    Returns:
        np.ndarray: sorted and reorder splines.
    """
    length_list = []
    for i in np.unique(coord[:, 0]):
        length_list.append(total_length(coord[np.where(coord[:, 0] == i)[0], 1:]))

    sorted_id = np.argsort(length_list)

    sorted_list = [coord[np.where(coord[:, 0] == i)[0], 1:] for i in sorted_id]
    sorted_list = [
        np.hstack((np.repeat(i, len(sorted_list[i])).reshape(-1, 1), sorted_list[i]))
        for i in range(len(sorted_list))
    ]

    return np.concatenate(sorted_list)


def cut_150_degree(segments_array: np.ndarray):
    """
    Cut segments based on angles between adjacent vectors.

    Given an array of line segments, this function calculates angles between
    adjacent vectors in each line. If the angle is less than or equal to 150
    degrees, the segment is cut into two new segments.

    Args:
    segments_array(np. ndarray): Array of line segments where the first column
    indicates the segment id and the remaining columns represent
    the coordinates of points.

    Args:
        segments_array:

    Returns:
        Tuple[bool, np.ndarray]: Indicates whether any segment was cut,
            and New array of cut segments.
    """

    cut_segments = []
    loop_ = False

    # Loop through unique segment IDs
    for i in np.unique(segments_array[:, 0]):
        pc_ = segments_array[np.where(segments_array[:, 0] == i)[0], 1:]

        angles_ = [180]

        # Calculate angles for each line segment
        for j in range(len(pc_) - 2):
            angles_.append(
                angle_between_vectors(
                    np.array(pc_[j]) - np.array(pc_[j + 1]),
                    np.array(pc_[j + 2]) - np.array(pc_[j + 1]),
                )
            )
        angles_.append(180)

        # Check if any angle is less than or equal to 150 degrees
        if len([id_ for id_, k in enumerate(angles_) if k <= 150]) > 0:
            loop_ = True

            # Find the minimum angle and cut the segment
            min_angle_idx = np.where(angles_ == np.min(angles_))[0][0]
            cut_segments.append(pc_[: min_angle_idx + 1, :])
            cut_segments.append(pc_[min_angle_idx + 1 :, :])
        else:
            cut_segments.append(pc_)

    # Filter out single-point segments
    cut_segments = [c for c in cut_segments if len(c) > 1]

    # Create the output array
    return loop_, np.vstack(
        [
            np.hstack((np.repeat(id_, len(c)).reshape(-1, 1), c))
            for id_, c in enumerate(cut_segments)
        ]
    )



#######################################################################
#  TARDIS - Transformer And Rapid Dimensionless Instance Segmentation #
#                                                                     #
#  New York Structural Biology Center                                 #
#  Simons Machine Learning Center                                     #
#                                                                     #
#  Robert Kiewisz, Tristan Bepler                                     #
#  MIT License 2021 - 2024                                            #
#######################################################################

from torch.nn import init

from tardis_em.utils.errors import TardisError


def weights_init_kaiming(m):
    """
    Kaiming weight and bias initialization.

    Args:
        m: CNN block.
    """
    class_name = m.__class__.__name__

    if class_name.find("Conv3d") != -1:
        init.kaiming_normal_(tensor=m.weight.data)
    elif class_name.find("Conv2d") != -1:
        init.kaiming_normal_(tensor=m.weight.data)
    elif class_name.find("BatchNorm") != -1:
        init.normal_(tensor=m.weight.data, mean=1.0, std=0.02)
        init.constant_(tensor=m.bias.data, val=0.0)
    elif class_name.find("GroupNorm") != -1:
        init.normal_(tensor=m.weight.data, mean=1.0, std=0.02)
        init.constant_(tensor=m.bias.data, val=0.0)


def init_weights(net, init_type="kaiming"):
    """
    Wrapper for network module initialization.

    Args:
        net: Network to initialized.
        init_type (str): type of initialization.

    Raises:
        NotImplementedError: _description_
    """
    if init_type == "kaiming":
        net.apply(weights_init_kaiming)
    else:
        TardisError(
            "140",
            "tardis_em/cnn/ini_weights.py",
            f"initialization method {init_type} is not implemented!",
        )



#######################################################################
#  TARDIS - Transformer And Rapid Dimensionless Instance Segmentation #
#                                                                     #
#  New York Structural Biology Center                                 #
#  Simons Machine Learning Center                                     #
#                                                                     #
#  Robert Kiewisz, Tristan Bepler                                     #
#  MIT License 2021 - 2024                                            #
#######################################################################

from typing import Union

import numpy as np

from tardis_em.utils.errors import TardisError


def get_n_params(model):
    pp = 0
    for p in list(model.parameters()):
        nn = 1
        for s in list(p.size()):
            nn = nn * s
        pp += nn
    return pp


class EarlyStopping:
    """
    Early stopping to stop the training when the loss does not improve after
    certain epochs.

    Args:
        patience (int): how many epochs to wait before stopping when loss is
               not improving.
        min_delta (int): minimum difference between new loss and old loss for
               new loss to be considered as an improvement.
    """

    def __init__(self, patience=10, min_delta=0):
        self.patience = patience
        self.min_delta = min_delta
        self.counter = 0
        self.best_loss = None
        self.early_stop = False

    def __call__(
        self, val_loss: Union[float, None] = None, f1_score: Union[float, None] = None
    ):
        if val_loss is None and f1_score is None:
            TardisError(
                "124",
                "tardis_em/utils/utils",
                "Validation loss or F1 score is missing in early stop!",
            )

        if val_loss is not None:
            if self.best_loss is None:
                self.best_loss = val_loss
            elif self.best_loss - val_loss > self.min_delta:
                self.best_loss = val_loss
                self.counter = 0  # Reset counter if validation loss improves
            elif self.best_loss - val_loss < self.min_delta:
                self.counter += 1

                if self.counter >= self.patience:
                    print("INFO: Early stopping")
                    self.early_stop = True
        elif f1_score is not None:
            if self.best_loss is None:
                self.best_loss = f1_score
            elif self.best_loss - f1_score < self.min_delta:
                self.best_loss = f1_score
                self.counter = 0  # Reset counter if validation loss improves
            elif self.best_loss - f1_score > self.min_delta:
                self.counter += 1

                if self.counter >= self.patience:
                    print("INFO: Early stopping")
                    self.early_stop = True


def check_uint8(image: np.ndarray):
    """
    Simple check for uint8 array.

    If array is not compliant, then data are examine for a type of data (binary
    mask, image with int8 etc.) and converted back to uint8.

    Args:
        image (np.ndarray): Image for evaluation.
    """
    if np.all(np.unique(image) == [0, 1]):
        # Binary mask image
        return image
    elif np.all(np.unique(image) == [0, 254]):
        # Multi-label mask
        return np.array(np.where(image > 1, 1, 0), dtype=np.uint8)
    elif len(np.unique(image) > 2):
        # Raw image data
        return image.astype(np.uint8)
    elif np.all(image == 0):
        # Image is empty
        return image



#######################################################################
#  TARDIS - Transformer And Rapid Dimensionless Instance Segmentation #
#                                                                     #
#  New York Structural Biology Center                                 #
#  Simons Machine Learning Center                                     #
#                                                                     #
#  Robert Kiewisz, Tristan Bepler                                     #
#  MIT License 2021 - 2024                                            #
#######################################################################
from datetime import datetime
from os.path import join
from typing import Union, List, Tuple
import numpy as np

from tardis_em.analysis.geometry_metrics import (
    length_list,
    curvature_list,
    intensity_list,
)

from tardis_em.utils.errors import TardisError

from tardis_em._version import version


def analise_filaments(
    data: Union[np.ndarray, List, Tuple], image: Union[np.ndarray, List, Tuple] = None
) -> tuple:
    """

    Args:
        data (np.ndarray, List, Tuple): Directory to file path
        image (np.ndarray, None): Optional image to compute intensity along predicted splines.
    """
    if isinstance(data, np.ndarray):
        data = [data]
    if image is not None:
        if isinstance(image, np.ndarray):
            image = [image]

    length, curvature, tortuosity = [], [], []
    avg_intensity, sum_intensity = [], []
    avg_length_intensity, sum_length_intensity = [], []

    for id_, d_ in enumerate(data):
        length.append(length_list(d_))

        curvature_, tortuosity_ = curvature_list(d_, tortuosity_=True, mean_=True)
        curvature.append(curvature_)
        tortuosity.append(tortuosity_)

        if image is not None:
            intensity = intensity_list(d_, image[id_])

            avg_intensity_ = [np.mean(i).item(0) for i in intensity]
            avg_intensity.append(avg_intensity_)

            avg_length_intensity_ = [i / l for l, i in zip(length[id_], avg_intensity_)]
            avg_length_intensity.append(avg_length_intensity_)

            sum_intensity_ = [np.sum(i).item(0) for i in intensity]
            sum_intensity.append(sum_intensity_)

            sum_length_intensity_ = [i / l for l, i in zip(length[id_], sum_intensity_)]
            sum_length_intensity.append(sum_length_intensity_)

    if image is not None:
        return (
            length,
            curvature,
            tortuosity,
            avg_intensity,
            avg_length_intensity,
            sum_intensity,
            sum_length_intensity,
        )
    else:
        return length, curvature, tortuosity, None, None, None, None


def save_analysis(
    names: Union[List, Tuple], analysis: Union[List, Tuple], px_=None, save: str = None
):

    length, curvature, tortuosity = analysis[0], analysis[1], analysis[2]
    avg_intensity, avg_length_intensity = analysis[3], analysis[4]
    sum_intensity, sum_length_intensity = analysis[5], analysis[6]

    rows = np.sum([len(i) for i in length]).item(0)
    analysis_file = np.zeros((rows, 10), dtype=object)
    analysis_file[0, :] = [
        "File_Name",
        "No. of Filament",
        "Pixel_Size [nm]",
        "Length [nm]",
        "Curvature [0-inf]",
        "Tortuosity [1-inf]",
        "Avg. Intensity [U]",
        "Avg. Intensity / Length [U/nm]",
        "Sum. Intensity [U]",
        "Avg. Intensity / Length [U/nm]",
    ]

    iter_ = 0
    for i in range(len(length)):  # Iterate throw every file
        for j in range(len(length[i])):
            analysis_file[iter_, :] = [
                names[i][:-21],
                str(j),
                str(1.0) if px_ is None else px_[i][j],
                str(length[i][j]),
                str(curvature[i][j]),
                str(tortuosity[i][j]),
                "" if avg_intensity[i] is None else str(avg_intensity[i][j]),
                (
                    ""
                    if avg_length_intensity[i] is None
                    else str(avg_length_intensity[i][j])
                ),
                "" if sum_intensity[i] is None else str(sum_intensity[i][j]),
                (
                    ""
                    if sum_length_intensity[i] is None
                    else str(sum_length_intensity[i][j])
                ),
            ]
            iter_ += 1

    if save is not None:
        date = datetime.now()
        tardis_version = version

        file_name = f"TARDIS_V{tardis_version}_{date.day}_{date.month}_{date.year}-{date.hour}_{date.minute}.csv"

        np.savetxt(
            join(save, file_name), analysis_file.astype(str), fmt="%s", delimiter=","
        )
        return
    return analysis_file


def analise_filaments_list(
    data: Union[List, Tuple],
    names_: Union[List, Tuple],
    path: str,
    images: Union[List, Tuple] = None,
    px_: Union[List, Tuple] = None,
):
    """
    Arge:
        data (list, tuple):
        names_ (list, tuple):
        path (str):
        images (list, tuple):
        px_ (list, tuple):
    """

    if images is not None:
        assert_ = len(data) == len(images)
    else:
        assert_ = len(data) > 0

    if not assert_:
        TardisError(
            id_="117",
            py="analysis.analysis:analise_filaments_list",
            desc="List of analysed files do not match or is 0",
            warning_=False,
        )

    if images is None:
        images = [None for _ in range(len(data))]

    save_analysis(names_, analise_filaments(data, images), px_=px_, save=path)



#######################################################################
#  TARDIS - Transformer And Rapid Dimensionless Instance Segmentation #
#                                                                     #
#  New York Structural Biology Center                                 #
#  Simons Machine Learning Center                                     #
#                                                                     #
#  Robert Kiewisz, Tristan Bepler                                     #
#  MIT License 2021 - 2024                                            #
#######################################################################
from typing import Optional, Tuple, Union

import numpy as np
import tifffile.tifffile as tiff
from sklearn.neighbors import KDTree, NearestNeighbors

from tardis_em.utils.errors import TardisError
from tardis_em.utils.load_data import ImportDataFromAmira
from tardis_em.utils.normalization import RescaleNormalize, SimpleNormalize


def preprocess_data(
    coord: Union[str, np.ndarray],
    image: Optional[str] = None,
    size: Optional[int] = None,
    include_label=True,
    normalization="simple",
) -> Union[Tuple[np.ndarray, np.ndarray], Tuple[np.ndarray, np.ndarray, np.ndarray]]:
    """
    Data augmentation function.

    Given any supported coordinate file, the function process it with optional image
    data. If image data is used, the image output is a list of flattened image patches
    of a specified size.
    Additionally, the graph output can be created.

    Args:
        coord (str, ndarray): Directory for the file containing coordinate data.
        image (str, None): Directory to the supported image file.
        size (int, None): Image patch size.
        include_label (bool): If True output coordinate array with label ids.
        normalization (str): Type of image normalization.

    Returns:
        Tuple[np.ndarray, np.ndarray]: Returns coordinates and optionally
        graph patch list.
    """

    """ Collect Coordinates [Length x Dimension] """
    if not isinstance(coord, np.ndarray):
        if coord[-4:] == ".csv":
            coord_label = np.genfromtxt(coord, delimiter=",")
            if str(coord_label[0, 0]) == "nan":
                coord_label = coord_label[1:, :]
        elif coord[-4:] == ".npy":
            coord_label = np.load(coord)
        elif coord[-3:] == ".am":
            if image is None:
                amira_import = ImportDataFromAmira(src_am=coord)
                coord_label = amira_import.get_segmented_points()
            else:
                if image.endswith(".am"):
                    amira_import = ImportDataFromAmira(src_am=coord, src_img=image)
                    coord_label = amira_import.get_segmented_points()
                else:
                    amira_import = ImportDataFromAmira(src_am=coord)
                    coord_label = amira_import.get_segmented_points()
    else:
        coord_label = coord.copy()

    if coord_label.shape[1] not in [3, 4]:
        TardisError(
            "",
            "tardis_em/dist_pytorch/dataset/augmentation.py",
            f"Coord file {coord} is without labels."
            "Expected dim to be in [3, 4] for 2D or 3D "
            f"coord but got {coord_label.shape[1]}",
        )

    """ Coordinates without labels """
    coords = coord_label[:, 1:]

    if size is None:
        size = 1

    if coords.shape[1] == 3:
        size = (size, size, size)
    else:
        size = (size, size)

    """ Collect Image Patches [Channels x Length] """
    # Normalize image between 0,1
    if image is not None:
        if normalization not in ["simple", "minmax", None]:
            TardisError(
                "124",
                "tardis_em/dist_pytorch/dataset/augmentation.py",
                f"Not implemented normalization. Given {normalization} "
                "But expected simple or minmax!",
            )

        if normalization == "simple":
            normalization = SimpleNormalize()
        elif normalization == "rescale":
            normalization = RescaleNormalize()

    if image is not None and coord[-3:] != ".am":
        # Crop images size around coordinates
        img_stack = tiff.imread(image)

        crop_tiff = Crop2D3D(
            image=img_stack, size=size, normalization=normalization
        )  # Z x Y x X

        # Load images in an array
        if len(size) == 2:
            img = np.zeros((len(coords), size[0] * size[1]))

            for i in range(img.shape[0]):
                point = coords[i]  # X x Y
                img[i, :] = np.array(crop_tiff(center_point=point)).flatten()

        elif len(size) == 3:
            img = np.zeros((len(coords), size[0] * size[1] * size[2]))

            for i in range(img.shape[0]):
                point = coords[i]  # X x Y x Z
                img[i, :] = np.array(crop_tiff(center_point=point)).flatten()

    elif image is not None and image.endswith(".am"):
        """Collect Image Patches for .am binary files"""
        img_stack, _ = amira_import.get_image()

        # Crop image around coordinates
        crop_tiff = Crop2D3D(
            image=img_stack, size=size, normalization=normalization
        )  # Z x Y x X

        # Load images patches into an array
        if len(size) == 2:
            img = np.zeros((len(coords), size[0] * size[1]))
        else:
            img = np.zeros((len(coords), size[0] * size[1] * size[2]))

        for i in range(img.shape[0]):
            point = coords[i]  # X x Y x Z
            img[i, :] = np.array(crop_tiff(center_point=point)).flatten()
    else:
        img = np.zeros(size)

    """ If not Include label build graph """
    if include_label:
        return coord_label, img
    else:
        graph_builder = BuildGraph()
        graph = graph_builder(coord=coord_label)

        return coords, img, graph


class BuildGraph:
    """
    GRAPH REPRESENTATION FROM 2D/3D COORDINATES

    The main class is to build a graph representation of any given point cloud based on
    the labeling information and optionally point distances.

    The BuildGraph class outputs a 2D array of a graph representation built for the
    filament-like structure which allows for only a maximum of 2 connections per node.
    Or for an object structure where the cap interaction per node limit was fixed as 4.
    The graph representation for a mesh-like object is computed by identifying
    all points in the class and searching for 4 KNN for each node inside the class.

    Args:
        K (int): Number of maximum connections per node.
    """

    def __init__(self, K=2, mesh=False):
        self.K = K
        self.mesh = mesh

    def __call__(self, coord: np.ndarray) -> np.ndarray:
        """
        Graph representation builder.

        Assuming the coordinate array is stored in a variable named 'coords',
        where the first column represents the class ID and the remaining three
        columns represent the XYZ coordinates.

        Args:
            coord (np.ndarray): A coordinate array of the shape (Nx[3, 4]).

        Returns:
            np.ndarray: Graph representation 2D array.
        """
        # extract the class ID and XYZ coordinates from the array
        class_id = coord[:, 0]
        xyz = coord[:, 1:]

        # build the connectivity matrix
        N = coord.shape[0]
        graph = np.zeros((N, N))
        if self.mesh:
            # build a NearestNeighbors object for efficient nearest neighbor search
            nn = NearestNeighbors(n_neighbors=self.K, algorithm="kd_tree").fit(xyz)

            # find the indices of the K-nearest neighbors for each point
            _, indices = nn.kneighbors(xyz)

            for i in range(N):
                for j in indices[i]:
                    if class_id[i] == class_id[j]:  # check class ID before adding edges
                        graph[i, j] = 1
                        # graph[j, i] = 1
        else:
            all_idx = np.unique(coord[:, 0])
            for i in all_idx:
                points_in_contour = np.where(coord[:, 0] == i)[0].tolist()

                for j in points_in_contour:
                    # Self-connection
                    graph[j, j] = 1

                    # First point in contour
                    if j == points_in_contour[0]:  # First point
                        if (j + 1) <= (len(coord) - 1):
                            graph[j, j + 1] = 1
                    # Last point
                    elif j == points_in_contour[len(points_in_contour) - 1]:
                        graph[j, j - 1] = 1
                    else:  # Point in the middle
                        graph[j, j + 1] = 1
                        graph[j, j - 1] = 1

                # Check Euclidean distance between fist and last point
                ends_distance = np.linalg.norm(
                    coord[points_in_contour[0]][1:] - coord[points_in_contour[-1]][1:]
                )

                # If < 2 nm pixel size, connect
                if ends_distance < 5:
                    graph[points_in_contour[0], points_in_contour[-1]] = 1
                    graph[points_in_contour[-1], points_in_contour[0]] = 1

        # Ensure self-connection
        np.fill_diagonal(graph, 1)

        return graph


class Crop2D3D:
    """
    2D/3D IMAGE CROPPING

    Center crop gave the image to the specified size.

    Args:
        image (np.ndarray): Image array.
        size (tuple): Uniform cropping size.
        normalization (class): Normalization type.
    """

    def __init__(self, image: np.ndarray, size: tuple, normalization):
        self.image = image
        self.size = size
        self.normalization = normalization

        if len(size) == 2:
            self.width, self.height = image.shape
            self.depth = None
        else:
            self.depth, self.width, self.height = image.shape

    @staticmethod
    def get_xyz_position(
        center_point: int, size: int, max_size: int
    ) -> Tuple[int, int]:
        """
        Given the center point, calculate the range to crop.

        Args:
            center_point (int): XYZ coordinate for center point.
            size (int): Crop size.
            max_size (int): Axis maximum size is used to calculate the offset.

        Returns:
            Tuple[int, int]: Min and max int values refer to the position on the axis.
        """
        x0 = center_point - (size / 2)
        x1 = center_point + (size / 2)

        """ Check if frame feet into image, if not shift it """
        if x0 < 0:
            move_by = abs(x0)
            x0 = x0 + move_by
            x1 = x1 + move_by

        if x1 > max_size:
            move_by = x1 - max_size
            x0 = x0 - move_by
            x1 = x1 - move_by

        return int(x0), int(x1)

    def __call__(self, center_point: tuple) -> np.ndarray:
        """
        Call for image cropping.

        Args:
            center_point (tuple): XYZ coordinates to crop image.

        Returns:
            np.ndarray: Cropped image patch.
        """
        if len(center_point) not in [2, 3]:
            TardisError(
                "113",
                "tardis_em/dist_pytorch/dataset/augmentation.py",
                "Given position for cropping is not 2D or 3D!. "
                f"Given {center_point}. But expected shape in [2, 3]!",
            )
        if len(center_point) != len(self.size):
            TardisError(
                "124",
                "tardis_em/dist_pytorch/dataset/augmentation.py",
                f"Given cropping shape {len(self.size)} is not compatible "
                f"with given cropping coordinates shape {len(center_point)}!",
            )

        if len(center_point) == 3:
            z0, z1 = self.get_xyz_position(
                center_point=center_point[-1], size=self.size[-1], max_size=self.depth
            )
            x0, x1 = self.get_xyz_position(
                center_point=center_point[0], size=self.size[0], max_size=self.height
            )
            y0, y1 = self.get_xyz_position(
                center_point=center_point[1], size=self.size[1], max_size=self.width
            )
            crop_img = self.image[z0:z1, y0:y1, x0:x1]

            if crop_img.shape != (self.size[-1], self.size[0], self.size[1]):
                crop_df = np.array(crop_img)
                shape = crop_img.shape

                crop_img = np.zeros((self.size[2], self.size[0], self.size[1]))
                crop_img[0 : shape[0], 0 : shape[1], 0 : shape[2]] = crop_df
        elif len(center_point) == 2:
            x0, x1 = self.get_xyz_position(
                center_point=center_point[0], size=self.size[0], max_size=self.height
            )
            y0, y1 = self.get_xyz_position(
                center_point=center_point[1], size=self.size[1], max_size=self.width
            )
            crop_img = self.image[y0:y1, x0:x1]

            if crop_img.shape != (self.size[0], self.size[1]):
                crop_df = np.array(crop_img)
                shape = crop_img.shape
                crop_img = np.zeros((self.size[0], self.size[1]))
                crop_img[0 : shape[0], 0 : shape[1]] = crop_df

        if self.normalization is not None:
            return self.normalization(crop_img)
        else:
            return crop_img


def upsample_pc(org_coord: np.ndarray, sampled_coord: np.ndarray):
    """
    upsample_pc _summary_

    Args:
        org_coord (np.ndarray): _description_
        sampled_coord (np.ndarray): _description_
    """
    # build a KDTree for efficient nearest neighbor search
    tree = KDTree(sampled_coord[:, 1:])

    # find the indices of the K-nearest neighbors for each point
    _, indices = tree.query(org_coord[:, 1:], k=2)

    # Take only first NN point
    indices = indices[:, 1, np.newaxis]
    indices = sampled_coord[indices, 0]

    return np.hstack((indices, org_coord[:, 1]))



#######################################################################
#  TARDIS - Transformer And Rapid Dimensionless Instance Segmentation #
#                                                                     #
#  New York Structural Biology Center                                 #
#  Simons Machine Learning Center                                     #
#                                                                     #
#  Robert Kiewisz, Tristan Bepler                                     #
#  MIT License 2021 - 2024                                            #
#######################################################################

from typing import Optional, Tuple, Union

import numpy as np
import torch
import torch.nn.functional as F
from PIL import Image
from torch.fft import fftn, ifftn, fftshift, ifftshift
from scipy.ndimage import gaussian_filter


def scale_image(
    scale: tuple,
    image: Optional[np.ndarray] = None,
    mask: Optional[np.ndarray] = None,
    device=torch.device("cpu"),
) -> Union[
    Tuple[np.ndarray, np.ndarray, int], Tuple[np.ndarray, int], Tuple[None, int]
]:
    """
    Scale image module using torch GPU interpolation

    Expect 2D (XY/YX), 3D (ZYX)

    Args:
        image (np.ndarray, Optional): image data
        mask (np.ndarray, Optional): Optional binary mask image data
        scale (tuple): scale value for image.
        device (torch.device):
    """
    dim = 1

    if len(scale) == 3:
        scale = tuple([scale[0], scale[1], scale[2]])
    else:
        scale = tuple([scale[0], scale[1]])

    if image is not None:
        image = image.astype(np.float32)
        image = np.ascontiguousarray(image)

        if not np.all(scale == image.shape):
            if scale[0] < image.shape[0]:  # Down-sampling
                image = linear_scaling(img=image, scale=scale, device=device)
            else:  # Up-sampling
                image = nn_scaling(
                    img=image,
                    scale=scale,
                    device=device,
                    gauss=True,
                )

    if mask is not None:
        mask = mask.astype(np.float32)
        mask = np.ascontiguousarray(mask)

        if not np.all(scale == mask.shape):
            mask = nn_scaling(img=mask, scale=scale)

    if image is not None and mask is not None:
        return image, mask, dim
    elif image is None and mask is not None:
        return mask, dim
    else:
        return image, dim


def nn_scaling(
    img: np.ndarray, scale: tuple, device=torch.device("cpu"), gauss=False
) -> np.ndarray:
    if gauss:
        filter_ = img.shape[0] / scale[0]  # Scaling factor
        img = gaussian_filter(img, filter_ / 2)

    img = torch.from_numpy(img)[None, None, ...].to(device)
    img = F.interpolate(img, size=scale, mode="nearest")[0, 0, ...]

    return img.cpu().detach().numpy()


def linear_scaling(
    img: np.ndarray, scale: tuple, device=torch.device("cpu")
) -> np.ndarray:
    """
    Scaling of 2D/3D array using trilinear method from pytorch

    Args:
        img: image array.
        scale: Scale array size.
        device (torch.device): Compute device

    Returns:
        no.ndarray: Up or Down scale 3D array.
    """
    if img.ndim == 3:
        current_depth, current_height, current_width = (
            img.shape[0],
            img.shape[1],
            img.shape[2],
        )
        final_depth, final_height, final_width = scale

        img = torch.from_numpy(img[None, None, :]).to(device).type(torch.float)

        # Calculate the number of steps required to reach the final scale by halving/doubling
        # Use logarithm base 2 since scaling is done by 2x at each step
        height_steps = int(abs(torch.log2(torch.tensor(final_height / current_height))))
        width_steps = int(abs(torch.log2(torch.tensor(final_width / current_width))))
        depth_steps = int(abs(torch.log2(torch.tensor(final_depth / current_depth))))

        # Perform scaling in steps
        for _ in range(max(height_steps, width_steps, depth_steps)):
            # Calculate intermediate scale
            new_height = int(
                current_height * (2 if current_height < final_height else 0.5)
            )
            new_width = int(current_width * (2 if current_width < final_width else 0.5))
            new_depth = int(current_depth * (2 if current_depth < final_depth else 0.5))

            # Stop if the desired scale is reached or exceeded
            if (
                new_height >= final_height
                and new_width >= final_width
                and new_depth >= final_depth
            ):
                break

            # Resize image
            img = F.interpolate(
                img,
                size=(new_height, new_width, new_depth),
                mode="trilinear",
                align_corners=False,
            )

            # Update current dimensions
            current_depth, current_height, current_width = (
                new_depth,
                new_height,
                new_width,
            )

            # Stop if the desired scale is reached or exceeded
            if (
                current_height >= final_height
                and current_width >= final_width
                and current_depth >= final_depth
            ):
                break

        img = F.interpolate(
            img,
            size=(scale[0], scale[1], scale[2]),
            mode="trilinear",
            align_corners=False,
        )
    else:
        current_height, current_width = img.shape[0], img.shape[1]
        final_height, final_width = scale

        img = torch.from_numpy(img[None, None, :]).to(device).type(torch.float)

        # Calculate the number of steps required to reach the final scale by halving/doubling
        # Use logarithm base 2 since scaling is done by 2x at each step
        height_steps = int(abs(torch.log2(torch.tensor(final_height / current_height))))
        width_steps = int(abs(torch.log2(torch.tensor(final_width / current_width))))

        # Perform scaling in steps
        if max(height_steps, width_steps) > 0:
            for _ in range(max(height_steps, width_steps)):
                # Calculate intermediate scale
                new_height = int(
                    current_height * (2 if current_height < final_height else 0.5)
                )
                new_width = int(
                    current_width * (2 if current_width < final_width else 0.5)
                )

                # Stop if the desired scale is reached or exceeded
                if new_height >= final_height and new_width >= final_width:
                    break

                # Resize image
                img = F.interpolate(
                    img,
                    size=(new_height, new_width),
                    mode="bilinear",
                    align_corners=False,
                )
                current_height, current_width = new_height, new_width

        img = F.interpolate(img, size=scale, mode="bilinear", align_corners=False)
    return img.detach().numpy()[0, 0, :]


def area_scaling(
    img: np.ndarray, scale: tuple, device=torch.device("cpu")
) -> np.ndarray:
    """
    Scaling of 3D array using area method from pytorch

    Args:
        img: 3D array.
        scale: Scale array size.
        device (torch.device): Compute device

    Returns:
        no.ndarray: Up or Down scale 3D array.
    """

    size_Z = [scale[0], img.shape[1], img.shape[2]]
    image_scale_Z = np.zeros(size_Z)

    # Scale Z axis
    for i in range(img.shape[2]):
        df_img = torch.from_numpy(img[:, :, i]).to(device).type(torch.float)

        image_scale_Z[:, :, i] = (
            F.interpolate(
                df_img[None, None, :], size=[int(s) for s in size_Z[:2]], mode="area"
            )
            .cpu()
            .detach()
            .numpy()[0, 0, :]
        )

    # Scale XY axis
    img = np.zeros(scale)
    for i in range(scale[0]):
        df_img = torch.from_numpy(image_scale_Z[i, :]).to(device).type(torch.float)
        img[i, :] = (
            F.interpolate(
                df_img[None, None, :], size=[int(s) for s in scale[1:]], mode="area"
            )
            .cpu()
            .detach()
            .numpy()[0, 0, :]
        )

    return img


def fourier_scaling(
    img: np.ndarray, scale: tuple, device=torch.device("cpu")
) -> np.ndarray:
    """
    Resize a 2D or 3D image using Fourier cropping.

    Parameters:
    img (np.ndarray): 2D or 3D array representing the image.
    scale (tuple): Desired shape (for 2D: (height, width), for 3D: (depth, height, width)).
    dtype (np.dtype): Desired output data type.
    device (torch.device): Torch device

    Returns:
        np.ndarray: Resized image in the desired data type.
    """
    if not isinstance(scale, tuple):
        scale = tuple(scale)

    # Get the current shape
    org_shape = img.shape

    # Calculate start indices for cropping or padding
    start_indices = [(cs - s) // 2 for cs, s in zip(org_shape, scale)]

    if not isinstance(img, torch.Tensor):
        img = torch.from_numpy(img)
    img.to(device)

    # Perform Fourier Transform
    f_image = fftn(img)

    # Shift the zero frequency component to the center
    f_image_shifted = fftshift(f_image)

    # Slicing logic for cropping or padding
    slices_current = [
        slice(max(0, start), max(0, start) + min(cs, s))
        for start, cs, s in zip(start_indices, org_shape, scale)
    ]
    slices_resized = [
        slice(max(0, -start), max(0, -start) + min(cs, s))
        for start, cs, s in zip(start_indices, org_shape, scale)
    ]

    resized_f_image_shifted = torch.zeros(scale, dtype=torch.complex64, device=device)
    resized_f_image_shifted[slices_resized] = f_image_shifted[slices_current]

    # Shift the zero frequency component back to the original position
    resized_f_image = ifftshift(resized_f_image_shifted)

    # Perform Inverse Fourier Transform
    resized_image = ifftn(resized_f_image)

    # Return the real part of the resized image
    return torch.abs(resized_image).cpu().detach().numpy()



#######################################################################
#  TARDIS - Transformer And Rapid Dimensionless Instance Segmentation #
#                                                                     #
#  New York Structural Biology Center                                 #
#  Simons Machine Learning Center                                     #
#                                                                     #
#  Robert Kiewisz, Tristan Bepler                                     #
#  MIT License 2021 - 2024                                            #
#######################################################################
from os import listdir
from os.path import isfile, join
from typing import Tuple, Union

import numpy as np

from tardis_em.dist_pytorch.utils.build_point_cloud import BuildPointCloud
from tardis_em.cnn.data_processing.draw_mask import draw_instances
from tardis_em.cnn.data_processing.trim import trim_with_stride
from tardis_em.utils.errors import TardisError
from tardis_em.utils.load_data import ImportDataFromAmira, load_image
from tardis_em.utils.logo import print_progress_bar, TardisLogo


def build_train_dataset(
    dataset_dir: str,
    circle_size: int,
    resize_pixel_size: Union[float, None],
    trim_xy: int,
    trim_z: int,
    benchmark=False,
    correct_pixel_size=None,
):
    """
    Module for building train datasets from compatible files.

    This module builds a train dataset from each file in the specified dir.
    Module working on the file as .tif/.mrc/.rec/.am and mask in image format
    of .csv and .am:
    - If mask is .csv, module expect image file in format of .tif/.mrc/.rec/.am
    - If mask is .am, module expect image file in format of .am

    For the given dir, module recognizes file, and then for each file couple
    (e.g. image and mask) if needed (e.g. mask is not .tif) it's build mask from
    the point cloud. For this module, expect that mask file (.csv, .am) contains
    coordinates values. Then for each point module draws 2D label.

    These image arrays are then scaled up or down to fit given pixel size, and
    finally arrays are trimmed with overlay (stride) for specific size.

    Files are saved in ./dir/train/imgs and ./dir/train/masks.

    Mask is computed as np.uint8 arrays and images as np.float64 normalized image
    value between 0 and 1.

    Args:
        dataset_dir (str): Directory with train test folders.
        circle_size (int): Size of the segmented object in A.
        resize_pixel_size (float, None): Pixel size for image resizing.
        trim_xy (int): Voxel size of output image in x and y dimension.
        trim_z (int): Voxel size of output image in z dimension.
        benchmark (bool): If True construct data for benchmark.
        correct_pixel_size (float, None): Optionally, correction for pixel size value.
    """
    """Setup"""
    # Activate Tardis progress bar
    tardis_progress = TardisLogo()
    tardis_progress(title="Data pre-processing for CNN")

    # Builder for point cloud
    b_pc = BuildPointCloud()

    clean_empty = not benchmark

    # All expected formats
    IMG_FORMATS = (".am", ".mrc", ".rec", ".map", ".tif")
    MASK_FORMATS = (
        ".CorrelationLines.am",
        "_mask.am",
        "_mask.mrc",
        "_mask.rec",
        "_mask.csv",
        "_mask.tif",
    )

    """Check what file are in the folder to build dataset"""
    img_list = [
        f
        for f in listdir(dataset_dir)
        if f.endswith(IMG_FORMATS) and not f.endswith(MASK_FORMATS)
    ]

    """For each image find matching mask, pre-process, trim and save"""
    img_counter = 0
    log_file = np.zeros((len(img_list) + 1, 9), dtype=object)
    log_file[0, :] = np.array(
        (
            "ID",
            "File_Name",
            "Pixel_Size",
            "Scale_Factor",
            "Mask_Type",
            "Image_Min",
            "Image_Max",
            "Total_Patches_No",
            "Saved_Patches_No",
        )
    )

    id_ = 1
    for i in img_list:
        """Update progress bar"""
        tardis_progress(
            title="Data pre-processing for CNN training",
            text_1="Building Training dataset:",
            text_2=f"Files: {i}",
            text_3="px: NaN",
            text_4="Scale: NaN",
            text_6="Image dtype: NaN",
            text_7=print_progress_bar(id_, len(img_list)),
        )

        log_file[id_, 0] = str(id_)
        np.savetxt(
            join(dataset_dir, "log.csv"), log_file.astype(str), fmt="%s", delimiter=","
        )

        """Get image directory and check if img is a file"""
        img_dir = join(dataset_dir, i)
        if not isfile(img_dir):
            # Store fail in the log file
            log_file = error_log_build_data(
                dir_=join(dataset_dir, "log.csv"), log_file=log_file, id_=id_, i=i
            )
            continue

        """Get matching mask file and check if maks is a file"""
        mask_prefix = 0
        mask_name = ""
        mask_dir = ""
        while not isfile(join(dataset_dir, mask_name)):
            if mask_prefix >= len(MASK_FORMATS):
                break

            mask_name = (
                i[:-3] + MASK_FORMATS[mask_prefix]
                if i.endswith(".am")
                else i[:-4] + MASK_FORMATS[mask_prefix]
            )

            mask_dir = join(dataset_dir, mask_name)
            mask_prefix += 1

        if not isfile(mask_dir):
            # Store fail in the log file
            log_file = error_log_build_data(
                dir_=join(dataset_dir, "log.csv"),
                log_file=log_file,
                id_=id_,
                i=i + "|" + mask_dir,
            )
            continue

        """Load files"""
        image, mask, pixel_size = load_img_mask_data(img_dir, mask_dir)
        if correct_pixel_size is not None:
            pixel_size = correct_pixel_size

        log_file[id_, 1] = str(i + "||" + mask_name)
        np.savetxt(
            join(dataset_dir, "log.csv"), log_file.astype(str), fmt="%s", delimiter=","
        )

        if image is None:
            continue

        if pixel_size is None:
            log_file = error_log_build_data(
                dir_=join(dataset_dir, "log.csv"),
                log_file=log_file,
                id_=id_,
                i=i + "||" + mask_dir,
            )

        """Calculate scale factor"""
        if resize_pixel_size is None:
            scale_factor = 1.0
        else:
            scale_factor = pixel_size / resize_pixel_size
            pixel_size = resize_pixel_size

        scale_shape = [int(i * scale_factor) for i in image.shape]

        log_file[id_, 2] = str(pixel_size)
        log_file[id_, 3] = str(scale_factor)
        np.savetxt(
            join(dataset_dir, "log.csv"), log_file.astype(str), fmt="%s", delimiter=","
        )

        """Update progress bar"""
        tardis_progress(
            title="Data pre-processing for CNN training",
            text_1="Building Training dataset:",
            text_2=f"Files: {i} {mask_name}",
            text_3=f"px: {pixel_size}",
            text_4=f"Scale: {round(scale_factor, 2)}",
            text_6=f"Image dtype: {image.dtype} min: {image.min()} max: {image.max()}",
            text_7=print_progress_bar(id_, len(img_list)),
        )

        """Draw mask for coord or process mask if needed"""
        # Detect coordinate mask array
        if mask.ndim == 2 and mask.shape[1] in [3, 4]:
            # Scale mask to correct pixel size
            mask[:, 1:] = mask[:, 1:] * scale_factor

            # Draw mask from coordinates
            mask = draw_instances(
                mask_size=scale_shape,
                coordinate=mask,
                pixel_size=pixel_size,
                circle_size=circle_size,
            )
            log_file[id_, 4] = "coord"
            np.savetxt(
                join(dataset_dir, "log.csv"),
                log_file.astype(str),
                fmt="%s",
                delimiter=",",
            )
        else:  # Detect an image mask array
            # Convert to binary
            mask = np.where(mask > 0, 1, 0).astype(np.uint8)

            if mask.min() != 0 and mask.max() != 1:
                TardisError(
                    id_="115",
                    py="tardis_em/cnn/data_processing.md/build_training_dataset",
                    desc=f"Mask min: {mask.min()}; max: {mask.max()} "
                    "but expected min: 0 and max: >1",
                )

            # mask borders
            if mask.ndim == 2:
                mask[:1, :] = 0
                mask[:, :1] = 0
                mask[-1:, :] = 0
                mask[:, -1:] = 0
            else:
                mask[:, 1, :] = 0
                mask[:, -1:, :] = 0
                mask[..., :1] = 0
                mask[..., -1:] = 0

            # if scale_factor != 1.0:
            #     pc = b_pc.build_point_cloud(image=mask, as_2d=True)
            #     pc = pc * scale_factor
            #     pc = pc.astype(np.uint32)
            #
            #     # Remove gaps from conversion to int
            #     gaps = [
            #         i
            #         for i in list(range(min(pc[:, 2]), max(pc[:, 2]) + 1))
            #         if i not in np.unique(pc[:, 2])
            #     ]
            #
            #     gaps_pc = []
            #     for j in gaps:
            #         pick_coord = pc[np.where(pc[:, 2] == j + 1)[0], :]
            #         if len(pick_coord) == 0:
            #             pick_coord = pc[np.where(pc[:, 2] == j - 1)[0], :]
            #
            #         if len(pick_coord) > 0:
            #             pick_coord[:, 2] = j
            #             gaps_pc.append(pick_coord)
            #     if len(gaps_pc) > 0:
            #         gaps_pc = np.concatenate(gaps_pc)
            #         pc = np.concatenate((pc, gaps_pc))
            #
            #     # Draw mask from coordinates
            #     mask = draw_instances(
            #         mask_size=scale_shape,
            #         coordinate=pc,
            #         label=False,
            #         pixel_size=pixel_size,
            #         circle_size=circle_size,
            #     )
            #     # Convert to binary
            #     mask = np.where(mask > 0, 1, 0).astype(np.uint8)

            log_file[id_, 4] = "mask"
            np.savetxt(
                join(dataset_dir, "log.csv"),
                log_file.astype(str),
                fmt="%s",
                delimiter=",",
            )

        """Update progress bar"""
        tardis_progress(
            title="Data pre-processing for CNN training",
            text_1="Building Training dataset:",
            text_2=f"Files: {i} {mask_name}",
            text_3=f"px: {pixel_size}",
            text_4=f"Scale: {round(scale_factor, 2)}",
            text_6=f"Image dtype: {image.dtype} min: {image.min()} max: {image.max()}",
            text_7=print_progress_bar(id_, len(img_list)),
        )

        """Normalize image histogram"""
        log_file[id_, 5] = str(image.min())
        log_file[id_, 6] = str(image.max())
        np.savetxt(
            join(dataset_dir, "log.csv"), log_file.astype(str), fmt="%s", delimiter=","
        )

        tardis_progress(
            title="Data pre-processing for CNN training",
            text_1="Building Training dataset:",
            text_2=f"Files: {i} {mask_name}",
            text_3=f"px: {pixel_size}",
            text_4=f"Scale: {round(scale_factor, 2)}",
            text_6=f"Image dtype: {image.dtype} min: {image.min()} max: {image.max()}",
            text_7=print_progress_bar(id_, len(img_list)),
        )

        """Voxelize Image and Mask"""
        count = trim_with_stride(
            image=image,
            mask=mask,
            scale=scale_shape,
            trim_size_xy=trim_xy,
            trim_size_z=trim_z,
            clean_empty=clean_empty,
            keep_if=100,
            output=join(dataset_dir, "train"),
            image_counter=img_counter,
            log=True,
            pixel_size=pixel_size,
        )
        img_counter += 1
        log_file[id_, 7] = str(count[0])
        log_file[id_, 8] = str(count[1])
        np.savetxt(
            join(dataset_dir, "log.csv"), log_file.astype(str), fmt="%s", delimiter=","
        )
        id_ += 1


def load_img_mask_data(
    image: str, mask: str
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """
    Define file format and load adequately image and mask/coordinate file.

    Expected combination are:
        - Amira (image) + Amira (coord)
        - Amira (image) + csv (coord)
        - Amira (image) + Amira (mask) or MRC/REC (mask)
        - Amira (image) + tif (mask)

        - MRC/REC(image) + Amira (coord)! Need to check if coordinate is not transformed!
        - MRC/REC(image) + csv (coord)
        - MRC/REC(image) + Amira (mask) or MRC/REC (mask)
        - MRC/REC (image) + tif (mask)

    Args:
        image (str): Directory address to the image file
        mask(str): Directory address to the mask/coordinate file

    Returns:
        Tuple[np.ndarray, np.ndarray, np.ndarray]: Ndarray of image, mask, and pixel size
    """
    coord = None
    img_px, mask_px = 1, 1

    """Load Amira or MRC/REC image"""
    if image.endswith((".mrc", ".rec", ".map")):  # Image is MRC/REC(image)
        # Load image
        image, img_px = load_image(image)
    elif image.endswith(".am"):  # Image is Amira (image)
        if mask.endswith(".CorrelationLines.am"):  # Found Amira (coord)
            importer = ImportDataFromAmira(src_am=mask, src_img=image)
            image, img_px = importer.get_image()

            coord = importer.get_segmented_points()  # [ID x X x Y x Z]

            mask_px = img_px
        else:  # Image is Amira (image)
            image, img_px = load_image(image)
    elif image.endswith((".tif", ".tiff")):
        image, img_px = load_image(image)

    """Load Amira or MRC/REC or csv mask"""
    # Find maska file and load
    if mask.endswith(("_mask.mrc", "_mask.rec")):  # Mask is MRC/REC (mask)
        mask, mask_px = load_image(mask)
    elif mask.endswith("_mask.am"):  # Mask is Amira (mask)
        mask, mask_px = load_image(mask)
    elif (
        mask.endswith(".CorrelationLines.am") and coord is None
    ):  # Mask is Amira (coord)
        importer = ImportDataFromAmira(src_am=mask)
        coord = importer.get_segmented_points()  # [ID x X x Y x Z]
        mask_px = img_px
        coord[:, 1:] = coord[:, 1:] // mask_px
    elif mask.endswith("_mask.csv"):  # Mask is csv (coord)
        coord = np.genfromtxt(mask, delimiter=",")  # [ID x X x Y x (Z)]
        if np.all(coord[0, :].astype(str) == "nan"):
            coord = coord[1:, :]

        mask_px = img_px
    elif mask.endswith("_mask.tif"):
        mask, _ = load_image(mask)
        mask_px = img_px

    if not img_px == mask_px:
        if not mask_px == 1.0 and not img_px > 1:
            img_px = None

    if coord is not None:
        return image, coord, img_px
    else:
        return image, mask, img_px


def error_log_build_data(
    dir_: str, log_file: np.ndarray, id_: int, i: str
) -> np.ndarray:
    """
    Update log file with error for data that could not be loaded

    Args:
        dir_ (str): Save directory.
        log_file (np.ndarray): Current log file list.
        id_ (int): Data id.
        i (Str): Data name.

    Returns:
        list: List of updated logfile
    """

    # Store fail in the log file
    log_file[id_ + 1, 0] = str(id)
    log_file[id_ + 1, 1] = str(i)
    log_file[id_ + 1, 2] = "NA"
    log_file[id_ + 1, 3] = "NA"
    np.savetxt(dir_, log_file, fmt="%s", delimiter=",")

    return log_file



#######################################################################
#  TARDIS - Transformer And Rapid Dimensionless Instance Segmentation #
#                                                                     #
#  New York Structural Biology Center                                 #
#  Simons Machine Learning Center                                     #
#                                                                     #
#  Robert Kiewisz, Tristan Bepler                                     #
#  MIT License 2021 - 2024                                            #
#######################################################################

import sys
from os import getcwd
from typing import Optional

import torch
from torch import optim

from tardis_em.cnn.cnn import build_cnn_network
from tardis_em.cnn.trainer import CNNTrainer
from tardis_em.cnn.utils.utils import check_model_dict
from tardis_em.utils.device import get_device
from tardis_em.utils.errors import TardisError
from tardis_em.utils.losses import *
from tardis_em.utils.trainer import ISR_LR

# Setting for stable release to turn off all debug APIs
torch.backends.cudnn.benchmark = True
torch.autograd.set_detect_anomaly(mode=False)
torch.autograd.profiler.profile(enabled=False)
torch.autograd.profiler.emit_nvtx(enabled=False)


def train_cnn(
    train_dataloader,
    test_dataloader,
    model_structure: dict,
    checkpoint: Optional[str] = None,
    loss_function="bce",
    learning_rate=1,
    learning_rate_scheduler=False,
    early_stop_rate=10,
    device="gpu",
    warmup=100,
    epochs=1000,
):
    """
    Wrapper for CNN models.

    Args:
        train_dataloader (torch.DataLoader): DataLoader with train dataset.
        test_dataloader (torch.DataLoader): DataLoader with test dataset.
        model_structure (dict): Dictionary with model setting.
        checkpoint (None, optional): Optional, CNN model checkpoint.
        loss_function (str): Type of loss function.
        learning_rate (float): Learning rate.
        learning_rate_scheduler (bool): If True, LR_scheduler is used with training.
        early_stop_rate (int): Define max. number of epoch's without improvements
        after which training is stopped.
        device (torch.device): Device on which model is trained.
        warmup (int): Number of warm-up steps.
        epochs (int): Max number of epoch's.
    """
    """Losses"""
    loss_functions = [
        AdaptiveDiceLoss,
        BCELoss,
        WBCELoss,
        BCEDiceLoss,
        CELoss,
        DiceLoss,
        ClDiceLoss,
        ClBCELoss,
        SigmoidFocalLoss,
        LaplacianEigenmapsLoss,
        BCEMSELoss,
    ]
    losses_f = {f.__name__: f() for f in loss_functions}

    """Check input variable"""
    if checkpoint is not None:
        save_train = torch.load(checkpoint, map_location=device)
        if "model_struct_dict" in save_train.keys():
            model_structure = save_train["model_struct_dict"]

    model_structure = check_model_dict(model_structure)
    if not isinstance(device, torch.device) and isinstance(device, str):
        device = get_device(device)

    """Build CNN model"""
    try:
        model = build_cnn_network(
            network_type=model_structure["cnn_type"],
            structure=model_structure,
            img_size=model_structure["img_size"],
            prediction=False,
        )
    except:
        TardisError(
            "14",
            "tardis_em/cnn/train.py",
            f"CNNModelError: Model type: {type} was not build correctly!",
        )
        sys.exit()

    """Build TARDIS progress bar output"""
    print_setting = [
        f"Training is started for {model_structure['cnn_type']}:",
        f"Local dir: {getcwd()}",
        f"Training for {model_structure['cnn_type']} with "
        f"No. of Layers: {model_structure['num_conv_layers']} with "
        f"{model_structure['in_channel']} input and "
        f"{model_structure['out_channel']} output channel",
        f"Layers are build of {model_structure['layer_components']} modules, "
        f"train on {model_structure['img_size']} pixel images, "
        f"with {model_structure['conv_scaler']} up/down sampling "
        "channel scaler.",
    ]

    """Optionally: Load checkpoint for retraining"""
    if checkpoint is not None:
        model.load_state_dict(save_train["model_state_dict"])
    model = model.to(device)

    """Define loss function for training"""
    loss_fn = losses_f["BCELoss"]
    if loss_function in losses_f:
        loss_fn = losses_f[loss_function]

    """Build training optimizer"""
    if learning_rate_scheduler:
        optimizer = optim.NAdam(
            params=model.parameters(),
            betas=(0.9, 0.999),
            lr=learning_rate,
            eps=1e-8,
            momentum_decay=4e-3,
        )
    else:
        optimizer = optim.NAdam(
            params=model.parameters(),
            lr=learning_rate,
            betas=(0.9, 0.999),
            eps=1e-8,
            momentum_decay=4e-3,
        )

    """Optionally: Build learning rate scheduler"""
    if learning_rate_scheduler:
        optimizer = ISR_LR(
            optimizer, lr_mul=learning_rate, warmup_steps=warmup, scale=1
        )

    """Optionally: Checkpoint model"""
    if checkpoint is not None:
        optimizer.load_state_dict(save_train["optimizer_state_dict"])
        del save_train

    """Build trainer"""
    train = CNNTrainer(
        model=model,
        structure=model_structure,
        device=device,
        criterion=loss_fn,
        optimizer=optimizer,
        print_setting=print_setting,
        training_DataLoader=train_dataloader,
        validation_DataLoader=test_dataloader,
        lr_scheduler=learning_rate_scheduler,
        epochs=epochs,
        early_stop_rate=early_stop_rate,
        checkpoint_name=model_structure["cnn_type"],
    )

    """Train"""
    train.run_trainer()



#######################################################################
#  TARDIS - Transformer And Rapid Dimensionless Instance Segmentation #
#                                                                     #
#  New York Structural Biology Center                                 #
#  Simons Machine Learning Center                                     #
#                                                                     #
#  Robert Kiewisz, Tristan Bepler                                     #
#  MIT License 2021 - 2024                                            #
#######################################################################

import io
import json
from os import makedirs, mkdir
from os.path import expanduser, isdir, isfile, join
from typing import Optional

import requests

from tardis_em.utils.errors import TardisError


def get_benchmark_aws() -> dict:
    """
    Retrieve the best benchmarking score for a given NN type

    Returns:
        dict: Dictionary with keys[network name] and values[list of scores]
    """
    network_benchmark = requests.get(
        "https://tardis-weigths.s3.dualstack.us-east-1.amazonaws.com/benchmark/best_scores.json",
        timeout=(5, None),
    )

    if network_benchmark.status_code == 200:
        network_benchmark = json.loads(network_benchmark.content.decode("utf-8"))

    return network_benchmark


def put_benchmark_aws(data: dict, network: Optional[str] = "", model=None) -> bool:
    """
    Upload new or update dictionary stored on S3

    Args:
        data (dict): Dictionary with network the best metrics
        network (Optional, str): Benchmarking network name [e.g., fnet_32_microtubules_id].
        model (Optional, str): Optional dictionary to model.

    Returns:
        bool: True if save correctly
    """
    r = requests.put(
        "https://tardis-weigths.s3.dualstack.us-east-1.amazonaws.com/"
        "benchmark/best_scores.json",
        json.dumps(data, indent=2, default=str),
        timeout=(5, None),
    )

    if model is not None and r.status_code == 200:
        with open(model, "rb") as data:
            r_m = requests.put(
                "https://tardis-weigths.s3.dualstack.us-east-1.amazonaws.com/"
                f"benchmark/models/{network}.pth",
                data=data,
                timeout=(5, None),
            )

        return r_m.status_code == 200
    return r.status_code == 200


def get_all_version_aws(network: str, subtype: str, model: str):
    r = requests.get(
        "https://tardis-weigths.s3.dualstack.us-east-1.amazonaws.com/",
    )
    r = r.content.decode("utf-8")
    all_version = [
        f[1:-3].split("/")[-1]
        for f in r.split("Key")
        if f.startswith(f">tardis_em/{network}_{subtype}/{model}/V_")
    ]
    return [v for v in all_version if v.startswith("V_")]


def get_model_aws(https: str):
    return requests.get(
        https,
        timeout=(5, None),
    )


def get_weights_aws(
    network: str,
    subtype: str,
    model: Optional[str] = None,
    version: Optional[int] = None,
):
    """
    Module to download pre-train weights from S3 AWS bucket.

    Model weight stored on S3 bucket with the naming convention
    network_subtype/model/model_weights.pth
    References.:
    - fnet_32/microtubules_3d/V_XX/model_weights.pth
    - dist_triang/microtubules_3d/V_XX/model_weights.pth

    Weights are stored in ~/.tardis_em with the same convention and .txt
    file with file header information to identified update status for local file
    if the network connection can be established.

    Args:
        network (str): Type of network for which weights are requested.
        subtype (str): Sub-name of the network or sub-parameter for the network.
        model (str): Additional dataset name used for the DIST.
        version (int): Version of the model, be defaults we fetch the newest model
    """
    ALL_MODELS = ["unet", "unet3plus", "fnet_attn", "dist"]
    ALL_SUBTYPE = ["16", "32", "64", "96", "128", "triang", "full"]
    CNN = ["unet", "unet3plus", "fnet", "fnet_attn"]
    CNN_DATASET = [
        "microtubules_tirf",
        "microtubules_3d",
        "microtubules_2d",
        "membrane_3d",
        "membrane_2d",
        "actin_3d",
    ]
    DIST_DATASET = ["microtubules", "s3dis", "membrane_2d", "2d", "3d"]

    """Chech dir"""
    if not isdir(join(expanduser("~"), ".tardis_em")):
        mkdir(join(expanduser("~"), ".tardis_em"))
    if not isdir(join(expanduser("~"), ".tardis_em", f"{network}_{subtype}")):
        mkdir(join(expanduser("~"), ".tardis_em", f"{network}_{subtype}"))
    if not isdir(
        join(expanduser("~"), ".tardis_em", f"{network}_{subtype}", f"{model}")
    ):
        mkdir(join(expanduser("~"), ".tardis_em", f"{network}_{subtype}", f"{model}"))

    """Get weights for CNN"""
    dir_ = join(expanduser("~"), ".tardis_em", f"{network}_{subtype}", f"{model}")

    if network not in ALL_MODELS:
        TardisError(
            "19",
            "tardis_em/utils/aws.py",
            f"Incorrect CNN network selected {network}_{subtype}",
        )
    if subtype not in ALL_SUBTYPE:
        TardisError(
            "19",
            "tardis_em/utils/aws.py",
            f"Incorrect CNN subtype selected {network}_{subtype}",
        )

    if network in CNN:
        if model not in CNN_DATASET:
            TardisError(
                "19",
                "tardis_em/utils/aws.py",
                f"Incorrect CNN model selected {model} but expected {CNN_DATASET}",
            )

    if network == "dist":
        if model not in DIST_DATASET:
            TardisError(
                "19",
                "tardis_em/utils/aws.py",
                f"Incorrect DIST model selected {model} but expected {DIST_DATASET}",
            )

    all_version = get_all_version_aws(network, subtype, model)

    if len(all_version) == 0:
        version = None
    else:
        if version is not None:
            version_assert = len([v for v in all_version if v == f"V_{version}"])
            if not version_assert == 1:
                TardisError(
                    "19", "tardis_em/utils/aws.py", f"{model} of V{version} not found"
                )
            else:
                version = f"V_{version_assert}"
        else:
            version = f"V_{max([int(v.split('_')[1]) for v in all_version])}"

    if aws_check_with_temp(model_name=[network, subtype, model, version]):
        print(f"Loaded temp weights for: {network}_{subtype} {version}...")

        if isfile(join(dir_, "model_weights.pth")):
            return join(dir_, "model_weights.pth")
        else:
            TardisError("19", "tardis_em/utils/aws.py", "No weights found")
    else:
        print(f"Downloading new weights for: {network}_{subtype} {version}...")

        if version is None:
            weight = get_model_aws(
                "https://tardis-weigths.s3.dualstack.us-east-1.amazonaws.com/tardis_em/"
                f"{network}_{subtype}/"
                f"{model}/model_weights.pth"
            )
        else:
            weight = get_model_aws(
                "https://tardis-weigths.s3.dualstack.us-east-1.amazonaws.com/tardis_em/"
                f"{network}_{subtype}/"
                f"{model}/{version}/model_weights.pth"
            )

        # Save weights
        open(join(dir_, "model_weights.pth"), "wb").write(weight.content)

        # Save header
        with open(join(dir_, "model_header.json"), "w") as f:
            json.dump(dict(weight.headers), f)

        """Save temp weights"""
        if not isdir(join(expanduser("~"), ".tardis_em")):
            mkdir(join(expanduser("~"), ".tardis_em"))

        if not isdir(dir_):
            makedirs(dir_)

        print(f"Pre-Trained model download from S3 and saved/updated in {dir_}")

        weight = weight.content
        if "AccessDenied" in str(weight[:100]):
            return join(dir_, "model_weights.pth")
        return io.BytesIO(weight)


def aws_check_with_temp(model_name: list) -> bool:
    """
    Module to check aws up-to data status.

    Quick check if the local file exists and is up-to data with aws server.

    Args:
        model_name (list): Name of the NN model.

    Returns:
        bool: If True, the local file is up to date.
    """
    """Check if temp dir exist"""
    if not isdir(join(expanduser("~"), ".tardis_em")):
        return False  # No weight, first Tardis run, download from aws

    """Check for stored file header in ~/.tardis_em/..."""
    if not isfile(
        join(
            expanduser("~"),
            ".tardis_em",
            f"{model_name[0]}_{model_name[1]}",
            f"{model_name[2]}",
            "model_weights.pth",
        )
    ):
        return False  # Define network was never used with tardis_em, download from aws
    else:
        if not isfile(
            join(
                expanduser("~"),
                ".tardis_em",
                f"{model_name[0]}_{model_name[1]}",
                f"{model_name[2]}",
                "model_header.json",
            )
        ):
            return False  # Weight found but no JSON, download from aws
        else:
            try:
                save = json.load(
                    open(
                        join(
                            expanduser("~"),
                            ".tardis_em",
                            f"{model_name[0]}_{model_name[1]}",
                            f"{model_name[2]}",
                            "model_header.json",
                        )
                    )
                )
            except:
                save = None

    """Compare stored file with file stored on aws"""
    if save is None:
        print("Network cannot be checked! Connect to the internet next time!")
        return False  # Error loading JSON, download from aws
    else:
        try:
            if model_name[3] == None:
                weight = requests.get(
                    "https://tardis-weigths.s3.dualstack.us-east-1.amazonaws.com/tardis_em/"
                    f"{model_name[0]}_{model_name[1]}/"
                    f"{model_name[2]}/model_weights.pth",
                    stream=True,
                    timeout=(5, None),
                )
                aws = dict(weight.headers)
            else:
                weight = requests.get(
                    "https://tardis-weigths.s3.dualstack.us-east-1.amazonaws.com/tardis_em/"
                    f"{model_name[0]}_{model_name[1]}/"
                    f"{model_name[2]}/{model_name[3]}/model_weights.pth",
                    stream=True,
                    timeout=(5, None),
                )
                aws = dict(weight.headers)
        except:
            print("Network cannot be checked! Connect to the internet next time!")
            return True  # Found saved weight but cannot connect to aws

    try:
        aws_data = aws["Last-Modified"]
    except KeyError:
        aws_data = aws["Date"]

    try:
        save_data = save["Last-Modified"]
    except KeyError:
        save_data = save["Date"]

    if save_data == aws_data:
        return True  # Up-to data weight, load from local dir
    else:
        return False  # There is a new version on aws, download from aws


def aws_check_pkg_with_temp() -> bool:
    """
    Module to check aws up-to data status for OTA-Update.

    Quick check if the local pkg file exists and is up-to data with aws server.

    Returns:
        bool: If True, the local file is up to date.
    """
    """Check if temp dir exist"""
    if not isdir(join(expanduser("~"), ".tardis_em")):
        return False  # No pkg, first Tardis run, download from aws

    """Check for stored file header in ~/.tardis_em/pkg_header.json."""
    if not isfile(
        join(
            expanduser("~"),
            ".tardis_em/tardis_em-x.x.x-py3-none-any.whl",
        )
    ):
        return False  # PKG was never download, download from aws
    else:
        if not isfile(
            join(
                expanduser("~"),
                ".tardis_em/pkg_header.json",
            )
        ):
            return False  # PKG found but no header, download from aws
        else:
            try:
                save = json.load(
                    open(
                        join(
                            expanduser("~"),
                            ".tardis_em/pkg_header.json",
                        )
                    )
                )
            except:
                save = None

    """Compare stored file with file stored on aws"""
    if save is None:
        print("Tardis_e pkg cannot be checked! Connect to the internet next time!")
        return False  # Error loading JSON, download from aws
    else:
        try:
            pkg = requests.get(
                "https://tardis-weigths.s3.dualstack.us-east-1.amazonaws.com/"
                "tardis_em/tardis_em-x.x.x-py3-none-any.whl",
                timeout=(5, None),
            )
            aws = dict(pkg.headers)
        except:
            print("Tardis_em pkg cannot be checked! Connect to the internet next time!")
            return True  # Found saved weight but cannot connect to aws

    try:
        aws_data = aws["Last-Modified"]
    except KeyError:
        aws_data = aws["Date"]

    try:
        save_data = save["Last-Modified"]
    except KeyError:
        save_data = save["Date"]

    if save_data == aws_data:
        return True  # Up-to data weight, load from local dir
    else:
        return False  # There is a new version on aws, download from aws



#######################################################################
#  TARDIS - Transformer And Rapid Dimensionless Instance Segmentation #
#                                                                     #
#  New York Structural Biology Center                                 #
#  Simons Machine Learning Center                                     #
#                                                                     #
#  Robert Kiewisz, Tristan Bepler                                     #
#  MIT License 2021 - 2024                                            #
#######################################################################

import numpy as np


def number_of_features_per_level(channel_scaler: int, num_levels: int) -> list:
    """
    Compute list of output channels for CNN.

    Features = channel_scaler * 2^k
        where:
        - k is layer number

    Args:
        channel_scaler (int): Number of initial input channels for CNN.
        num_levels (int): Number of channels from max_number_of_conv_layer().

    Returns:
        list: List of output channels.
    """
    return [channel_scaler * 2**k for k in range(num_levels)]


def max_number_of_conv_layer(
    img=None,
    input_volume=64,
    max_out=8,
    kernel_size=3,
    padding=1,
    stride=1,
    pool_size=2,
    pool_stride=2,
    first_max_pool=False,
) -> int:
    """
    Calculate maximum possible number of layers given image size.

    Based on the torch input automatically select number of convolution blocks,
    based on a standard settings.

        I = [(W - K + 2*P) / S] + 1
        Out_size = [(I - F) / S] + 1
        - W is the input volume
        - K is the Kernel size
        - P is the padding
        - S is the stride
        - F is the max pooling kernel size

    Args:
        img (np.ndarray, Optional): Tensor data from which size of is calculated.
        input_volume (int): Size of the multiplayer for the convolution.
        max_out (int): Maximal output dimension after max_pooling.
        kernel_size (int): Kernel size for the convolution.
        padding (int): Padding size for the convolution blocks.
        stride: (int) Stride for the convolution blocks.
        pool_size (int) Max Pooling kernel size.
        pool_stride (int): Max Pooling stride size.
        first_max_pool (bool): If first CNN block has max pooling.

    Returns:
        int: Maximum number of CNN layers.
    """
    if img is not None:
        # In case of anisotropic image select smallest size
        # to calculate max conv layers
        if len(img.shape) == 5:
            size = np.array(list(img.shape[2:]))
        elif len(img.shape) == 4:
            size = np.array(list(img.shape[1:]))
        else:
            size = np.array(list(img.shape))

        size = size[size.argmin()]
        input_volume = size

    max_layers = 0

    while input_volume >= max_out:
        img_window = ((input_volume - kernel_size + 2 * padding) / stride) + 1
        input_volume = ((img_window - pool_size) / pool_stride) + 1
        max_layers += 1

    if first_max_pool:
        max_layers -= 1

    return max_layers


def normalize_image(image: np.ndarray) -> np.ndarray:
    """
    Simple image data normalizer between 0,1.

    Args:
        image: Image data set.

    Returns:
        np.ndarray: Normalized image between 0 and 1 values.
    """
    image_min = np.min(image)
    image_max = np.max(image)

    if image_min == 0 and image_max == 1:
        return image

    if image_min == 0:
        image = np.where(image > image_min, 1, 0)
    elif image_max == 0:
        image = np.where(image < image_max, 1, 0)

    return image


def check_model_dict(model_dict: dict) -> dict:
    """
    Check and rebuild model structure dictionary to ensure back-compatibility.

    Args:
        model_dict (dict): Model structure dictionary.

    Returns:
        dict: Standardize model structure dictionary.
    """
    new_dict = {}

    for key, value in model_dict.items():
        if key.endswith("type"):
            new_dict["cnn_type"] = value
        if key.endswith("cation"):
            new_dict["classification"] = value
        if key.endswith("_in") or key.startswith("in_"):
            new_dict["in_channel"] = value
        if key.endswith("_out") or key.startswith("out_"):
            new_dict["out_channel"] = value
        if key.endswith("size"):
            new_dict["img_size"] = value
        if key.endswith("dropout"):
            new_dict["dropout"] = value
        if key.endswith("layers"):
            new_dict["num_conv_layers"] = value
        if key.endswith("scaler") or key.endswith("multiplayer"):
            new_dict["conv_scaler"] = value
        if key.endswith("v_kernel"):
            new_dict["conv_kernel"] = value
        if key.endswith("padding"):
            new_dict["conv_padding"] = value
        if key.endswith("l_kernel"):
            new_dict["maxpool_kernel"] = value
        if key.endswith("components"):
            new_dict["layer_components"] = value
        if key.endswith("features"):
            new_dict["attn_features"] = value
        if key.endswith("group"):
            new_dict["num_group"] = value

    return new_dict



#######################################################################
#  TARDIS - Transformer And Rapid Dimensionless Instance Segmentation #
#                                                                     #
#  New York Structural Biology Center                                 #
#  Simons Machine Learning Center                                     #
#                                                                     #
#  Robert Kiewisz, Tristan Bepler                                     #
#  MIT License 2021 - 2024                                            #
#######################################################################
import numpy as np
import os
from os.path import join
import tardis_em.utils.load_data as Loader
import tardis_em.utils.export_data as Exporter
import tifffile.tifffile as tiff

array_formats = (".csv", ".npy")


def tardis_helper(func: str, dir_=None, px: float = None):
    function = func.split("_")  # Expect e.g: csv_am
    in_files = [f for f in os.listdir(dir_) if f.endswith(function[0])]

    for i in in_files:
        # Detect images or coord arrays
        if i.endswith(array_formats):
            # Convert coordinates
            if function[0] == "csv":
                coord = np.genfromtxt(
                    join(dir_, i), delimiter=",", skip_header=1, dtype=np.float32
                )
            elif function[0] == "am":
                try:
                    am_import = Loader.ImportDataFromAmira(os.path.join(dir_, i))
                    if am_import is not None:
                        am_import = am_import.get_segmented_points()
                except:
                    continue
            elif function[0] == "npy":
                coord = np.load(join(dir_, i))

            if isinstance((coord[0, 0]), int) or isinstance((coord[0, 0]), float):
                if coord.shape[1] == 3:  # add 0 Z dimension
                    coord = np.vstack((coord, np.repeat(0, len(coord)).reshape(-1, 1)))

            if function[1] == "csv":
                np.savetxt(
                    join(dir_, i[: -len(function[0])] + "csv"),
                    coord,
                    delimiter=",",
                )
            elif function[1] == "am":
                Exporter.NumpyToAmira().export_amira(
                    join(dir_, i[: -len(function[0])] + "am"), coord
                )
        else:  # Convert image file
            image, px = Loader.load_image(join(dir_, i))

            if function[0] == "mrc":
                header = Loader.mrc_read_header(dir_)
            else:
                header = None

            if function[1] == "mrc":
                Exporter.to_mrc(
                    image, px, join(dir_, i[: -len(function[0])] + "mrc"), header
                )
            elif function[1] == "am":
                Exporter.to_am(image, px, join(dir_, i[: -len(function[0])] + "mrc"))
            elif function[1] == "tif" or function[1] == "tiff":
                tiff.imwrite(join(dir_, i[: -len(function[0])] + "mrc"), image)



#######################################################################
#  TARDIS - Transformer And Rapid Dimensionless Instance Segmentation #
#                                                                     #
#  New York Structural Biology Center                                 #
#  Simons Machine Learning Center                                     #
#                                                                     #
#  Robert Kiewisz, Tristan Bepler                                     #
#  MIT License 2021 - 2024                                            #
#######################################################################
import numpy as np
import scipy
import torch

from skimage import exposure


def adaptive_threshold(img: np.ndarray):
    if img.ndim == 3:
        std_ = np.std(img, axis=(1, 2), keepdims=True)
        mean_ = np.mean(img, axis=(1, 2), keepdims=True)
    else:
        std_ = np.std(img)
        mean_ = np.mean(img)

    # Perform the thresholding in a vectorized manner
    return (img >= (mean_ / std_)).astype(np.uint8)


class SimpleNormalize:
    """
    SIMPLE IMAGE NORMALIZATION

    Take int8-int32 image file with 0 - 255 value. All image value are spread
    between 0 - 1.
    """

    def __call__(self, x: np.ndarray) -> np.ndarray:
        """
        Call for image normalization.

        Args:
            x (np.ndarray): Image array.

        Returns:
            np.ndarray: Normalized image.
        """
        if x.dtype == np.uint8:
            x = x / 255
        elif x.dtype == np.int8:
            x = (x + 128) / 255
        elif x.dtype == np.uint16:
            x = x / 65535
        elif x.dtype == np.int16:
            x = (x + 32768) / 65535
        elif x.dtype == np.uint32:
            x = x / 4294967295
        else:
            x = (x + 2147483648) / 4294967295

        return x.astype(np.float32)


class MinMaxNormalize:
    """
    IMAGE NORMALIZATION BETWEEN MIN AND MAX VALUE
    """

    def __call__(self, x: np.ndarray) -> np.ndarray:
        """
        Call for normalization.

        Args:
            x (np.ndarray): Image data.

        Returns:
            np.ndarray: Normalized array.
        """
        MIN = np.min(x)
        MAX = np.max(x)

        if MAX <= 0:
            x = x + abs(MIN)
        x = (x - MIN) / (MAX - MIN)

        return x.astype(np.float32)


class MeanStdNormalize:
    """
    IMAGE NORMALIZATION BASED ON MEAN AND STD
    """

    def __call__(self, x: np.ndarray) -> np.ndarray:
        """
        Call for standardization.

        Args:
            x (np.ndarray): Image data.

        Returns:
            np.ndarray: Standardized array.
        """
        MEAN = float(x.mean())
        STD = float(x.std())

        x = (x - MEAN) / STD
        return x.astype(np.float32)


class RescaleNormalize:
    """
    NORMALIZE IMAGE VALUE USING Skimage

    Rescale intensity with top% and bottom% percentiles as default

    Args:
        clip_range: Histogram percentiles range crop.
    """

    def __init__(self, clip_range=(2, 98)):
        self.range = clip_range

    def __call__(self, x: np.ndarray) -> np.ndarray:
        """
        Call for normalization.

        Args:
            x (np.ndarray): Image or label mask.

        Returns:
            np.ndarray: Normalized array.
        """
        p2, p98 = np.percentile(x, self.range)
        if x.dtype == np.uint8:
            if p98 >= 250:
                p98 = 256
            if p2 <= 5:
                p2 = 0

        return exposure.rescale_intensity(x, in_range=(p2, p98))


class FFTNormalize:
    def __init__(
        self,
        method="affine",
        alpha=900,
        beta_=1,
        num_iters=100,
        sample=1,
        use_cuda=False,
    ):
        self.method = method
        self.alpha = alpha
        self.beta = beta_
        self.num_iters = num_iters
        self.sample = sample
        self.cuda = use_cuda

        self.mean_std = MeanStdNormalize()

    @staticmethod
    def gmm_fit(
        x,
        pi=0.5,
        split=None,
        alpha=0.5,
        beta_=0.5,
        scale=1.0,
        tol=1e-3,
        num_iters=100,
        share_var=True,
        verbose=False,
    ):
        # fit 2-component GMM
        # put a beta distribution prior on pi

        mu = torch.mean(x)
        pi = torch.as_tensor(pi)

        # split into everything > and everything <= pi pixel value
        # for assigning initial parameters
        if split is None:
            split = np.quantile(x.cpu().numpy(), 1 - pi)
        mask = x <= split

        p0 = mask.float()
        p1 = 1 - p0

        mu0 = mu
        s = torch.sum(p0)
        if s > 0:
            mu0 = torch.sum(x * p0) / s

        mu1 = mu
        s = torch.sum(p1)
        if s > 0:
            mu1 = torch.sum(x * p1) / s

        if share_var:
            var = torch.mean(p0 * (x - mu0) ** 2 + p1 * (x - mu1) ** 2)
            var0 = var
            var1 = var
        else:
            var0 = torch.sum(p0 * (x - mu0) ** 2) / torch.sum(p0)
            var1 = torch.sum(p1 * (x - mu1) ** 2) / torch.sum(p1)

        # first, calculate p(k | x, mu, var, pi)
        log_p0 = (
            -((x - mu0) ** 2) / 2 / var0
            - 0.5 * torch.log(2 * np.pi * var0)
            + torch.log1p(-pi)
        )
        log_p1 = (
            -((x - mu1) ** 2) / 2 / var1
            - 0.5 * torch.log(2 * np.pi * var1)
            + torch.log(pi)
        )

        ma = torch.max(log_p0, log_p1)
        Z = ma + torch.log(torch.exp(log_p0 - ma) + torch.exp(log_p1 - ma))

        # the probability of the data is
        logp = scale * torch.sum(Z) + scipy.stats.beta.logpdf(
            pi.cpu().numpy(), alpha, beta_
        )
        logp_cur = logp

        for it in range(1, num_iters + 1):
            # calculate the assignments
            p0 = torch.exp(log_p0 - Z)
            p1 = torch.exp(log_p1 - Z)

            # now, update distribution parameters
            s = torch.sum(p1)
            a = alpha + s
            b = beta_ + p1.numel() - s
            pi = (a - 1) / (a + b - 2)  # MAP estimate of pi

            mu0 = mu
            s = torch.sum(p0)
            if s > 0:
                mu0 = torch.sum(x * p0) / s

            mu1 = mu
            s = torch.sum(p1)
            if s > 0:
                mu1 = torch.sum(x * p1) / s

            if share_var:
                var = torch.mean(p0 * (x - mu0) ** 2 + p1 * (x - mu1) ** 2)
                var0 = var
                var1 = var
            else:
                var0 = torch.sum(p0 * (x - mu0) ** 2) / torch.sum(p0)
                var1 = torch.sum(p1 * (x - mu1) ** 2) / torch.sum(p1)

            # recalculate likelihood p(k | x, mu, var, pi)
            log_p0 = (
                -((x - mu0) ** 2) / 2 / var0
                - 0.5 * torch.log(2 * np.pi * var0)
                + torch.log1p(-pi)
            )
            log_p1 = (
                -((x - mu1) ** 2) / 2 / var1
                - 0.5 * torch.log(2 * np.pi * var1)
                + torch.log(pi)
            )

            ma = torch.max(log_p0, log_p1)
            Z = ma + torch.log(torch.exp(log_p0 - ma) + torch.exp(log_p1 - ma))

            logp = scale * torch.sum(Z) + scipy.stats.beta.logpdf(
                pi.cpu().numpy(), alpha, beta_
            )

            if verbose:
                print(it, logp)

            # check for termination
            if logp - logp_cur <= tol:
                break  # done
            logp_cur = logp

        return logp, mu0, mu1, var1

    def norm_fit(self, x, alpha=900, beta_=1, scale=1.0, num_iters=100, use_cuda=False):
        # try multiple initializations of pi
        pis = np.array([0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9, 0.95, 0.98, 1])
        splits = np.quantile(x, 1 - pis)

        logps = np.zeros(len(pis))
        mus = np.zeros(len(pis))
        stds = np.zeros(len(pis))

        x = torch.from_numpy(x)
        if use_cuda:
            x = x.cuda()

        for i in range(len(pis)):
            pi = float(pis[i])
            split = splits[i]
            if pi == 1:  # single component model
                mu = x.mean()
                var = x.var()
                logp = scale * torch.sum(
                    -((x - mu) ** 2) / 2 / var - 0.5 * torch.log(2 * np.pi * var)
                ) + scipy.stats.beta.pdf(1, alpha, beta_)
            else:
                logp, mu0, mu, var = self.gmm_fit(
                    x,
                    pi=pi,
                    split=split,
                    alpha=alpha,
                    beta_=beta_,
                    scale=scale,
                    num_iters=num_iters,
                )

            logps[i] = logp.item()
            mus[i] = mu.item()
            stds[i] = np.sqrt(var.item())

        # select normalization parameters with maximum logp
        i = np.argmax(logps)
        return mus[i], stds[i]

    def __call__(self, x: np.ndarray):
        if self.method == "affine":
            return self.mean_std(x)

        if self.sample > 1:
            n = int(np.round(x.size / self.sample))
            scale = x.size / n
            x_sample = np.random.choice(x.flatten(), size=n, replace=False)

            mu, std = self.norm_fit(
                x_sample,
                alpha=self.alpha,
                beta_=self.beta,
                scale=scale,
                num_iters=self.num_iters,
                use_cuda=self.cuda,
            )
        else:
            mu, std = self.norm_fit(
                x,
                alpha=self.alpha,
                beta_=self.beta,
                scale=1,
                num_iters=self.num_iters,
                use_cuda=self.cuda,
            )

        x = (x - mu) / std
        x = x.astype(np.float32)

        return x
